@@ -1,13 +1,36 @@
-/*	$OpenBSD: history.c,v 1.36 2008/05/20 00:30:30 fgsch Exp $	*/
-/*	$OpenBSD: trap.c,v 1.22 2005/03/30 17:16:37 deraadt Exp $	*/
+/*	$OpenBSD: history.c,v 1.39 2010/05/19 17:36:08 jasper Exp $	*/
+/*	$OpenBSD: trap.c,v 1.23 2010/05/19 17:36:08 jasper Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
+#if HAVE_PERSISTENT_HISTORY
+#include <sys/file.h>
+#endif
 
-__RCSID("$MirOS: src/bin/mksh/histrap.c,v 1.67 2008/07/12 17:16:03 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/histrap.c,v 1.97 2010/07/17 22:09:35 tg Exp $");
 
 /*-
  * MirOS: This is the default mapping type, and need not be specified.
- * IRIX doesn’t have this constant.
+ * IRIX doesn't have this constant.
  */
 #ifndef MAP_FILE
 #define MAP_FILE	0
@@ -28,7 +51,7 @@ static int sprinkle(int);
 
 static int hist_execute(char *);
 static int hist_replace(char **, const char *, const char *, int);
-static char **hist_get(const char *, int, int);
+static char **hist_get(const char *, bool, bool);
 static char **hist_get_oldest(void);
 static void histbackup(void);
 
@@ -56,7 +79,7 @@ c_fc(const char **wp)
 
 	if (!Flag(FTALKING_I)) {
 		bi_errorf("history functions not available");
-		return 1;
+		return (1);
 	}
 
 	while ((optc = ksh_getopt(wp, &builtin_opt,
@@ -73,7 +96,7 @@ c_fc(const char **wp)
 				memcpy(editor + len, " $_", 4);
 			}
 			break;
-		case 'g': /* non-at&t ksh */
+		case 'g': /* non-AT&T ksh */
 			gflag++;
 			break;
 		case 'l':
@@ -85,10 +108,10 @@ c_fc(const char **wp)
 		case 'r':
 			rflag++;
 			break;
-		case 's':	/* posix version of -e - */
+		case 's':	/* POSIX version of -e - */
 			sflag++;
 			break;
-		  /* kludge city - accept -num as -- -num (kind of) */
+		/* kludge city - accept -num as -- -num (kind of) */
 		case '0': case '1': case '2': case '3': case '4':
 		case '5': case '6': case '7': case '8': case '9':
 			p = shf_smprintf("-%c%s",
@@ -99,11 +122,11 @@ c_fc(const char **wp)
 				last = p;
 			else {
 				bi_errorf("too many arguments");
-				return 1;
+				return (1);
 			}
 			break;
 		case '?':
-			return 1;
+			return (1);
 		}
 	wp += builtin_opt.optind;
 
@@ -113,12 +136,12 @@ c_fc(const char **wp)
 
 		if (editor || lflag || nflag || rflag) {
 			bi_errorf("can't use -e, -l, -n, -r with -s (-e -)");
-			return 1;
+			return (1);
 		}
 
 		/* Check for pattern replacement argument */
 		if (*wp && **wp && (p = cstrchr(*wp + 1, '='))) {
-			pat = str_save(*wp, ATEMP);
+			strdupx(pat, *wp, ATEMP);
 			rep = pat + (p - *wp);
 			*rep++ = '\0';
 			wp++;
@@ -128,19 +151,19 @@ c_fc(const char **wp)
 			wp++;
 		if (last || *wp) {
 			bi_errorf("too many arguments");
-			return 1;
+			return (1);
 		}
 
 		hp = first ? hist_get(first, false, false) :
 		    hist_get_newest(false);
 		if (!hp)
-			return 1;
-		return hist_replace(hp, pat, rep, gflag);
+			return (1);
+		return (hist_replace(hp, pat, rep, gflag));
 	}
 
 	if (editor && (lflag || nflag)) {
 		bi_errorf("can't use -l, -n with -e");
-		return 1;
+		return (1);
 	}
 
 	if (!first && (first = *wp))
@@ -149,28 +172,28 @@ c_fc(const char **wp)
 		wp++;
 	if (*wp) {
 		bi_errorf("too many arguments");
-		return 1;
+		return (1);
 	}
 	if (!first) {
 		hfirst = lflag ? hist_get("-16", true, true) :
 		    hist_get_newest(false);
 		if (!hfirst)
-			return 1;
+			return (1);
 		/* can't fail if hfirst didn't fail */
 		hlast = hist_get_newest(false);
 	} else {
 		/* POSIX says not an error if first/last out of bounds
-		 * when range is specified; at&t ksh and pdksh allow out of
+		 * when range is specified; AT&T ksh and pdksh allow out of
 		 * bounds for -l as well.
 		 */
 		hfirst = hist_get(first, (lflag || last) ? true : false,
 		    lflag ? true : false);
 		if (!hfirst)
-			return 1;
+			return (1);
 		hlast = last ? hist_get(last, true, lflag ? true : false) :
 		    (lflag ? hist_get_newest(false) : hfirst);
 		if (!hlast)
-			return 1;
+			return (1);
 	}
 	if (hfirst > hlast) {
 		char **temp;
@@ -200,7 +223,7 @@ c_fc(const char **wp)
 			shf_fprintf(shl_stdout, "%s\n", s);
 		}
 		shf_flush(shl_stdout);
-		return 0;
+		return (0);
 	}
 
 	/* Run editor on selected lines, then run resulting commands */
@@ -209,14 +232,14 @@ c_fc(const char **wp)
 	if (!(shf = tf->shf)) {
 		bi_errorf("cannot create temp file %s - %s",
 		    tf->name, strerror(errno));
-		return 1;
+		return (1);
 	}
 	for (hp = rflag ? hlast : hfirst;
 	    hp >= hfirst && hp <= hlast; hp += rflag ? -1 : 1)
 		shf_fprintf(shf, "%s\n", *hp);
 	if (shf_close(shf) == EOF) {
 		bi_errorf("error writing temporary file - %s", strerror(errno));
-		return 1;
+		return (1);
 	}
 
 	/* Ignore setstr errors here (arbitrary) */
@@ -227,10 +250,10 @@ c_fc(const char **wp)
 		Source *sold = source;
 		int ret;
 
-		ret = command(editor ? editor : "${FCEDIT:-/bin/ed} $_");
+		ret = command(editor ? editor : "${FCEDIT:-/bin/ed} $_", 0);
 		source = sold;
 		if (ret)
-			return ret;
+			return (ret);
 	}
 
 	{
@@ -241,7 +264,7 @@ c_fc(const char **wp)
 
 		if (!(shf = shf_open(tf->name, O_RDONLY, 0, 0))) {
 			bi_errorf("cannot open temp file %s", tf->name);
-			return 1;
+			return (1);
 		}
 
 		n = stat(tf->name, &statb) < 0 ? 128 : statb.st_size + 1;
@@ -255,12 +278,12 @@ c_fc(const char **wp)
 			bi_errorf("error reading temp file %s - %s",
 			    tf->name, strerror(shf_errno(shf)));
 			shf_close(shf);
-			return 1;
+			return (1);
 		}
 		shf_close(shf);
 		*xp = '\0';
 		strip_nuls(Xstring(xs, xp), Xlength(xs, xp));
-		return hist_execute(Xstring(xs, xp));
+		return (hist_execute(Xstring(xs, xp)));
 	}
 }
 
@@ -280,24 +303,25 @@ hist_execute(char *cmd)
 			if (!*q) /* ignore trailing newline */
 				q = NULL;
 		}
-		histsave(++(hist_source->line), p, 1);
+		histsave(&hist_source->line, p, true, true);
 
 		shellf("%s\n", p); /* POSIX doesn't say this is done... */
-		if ((p = q)) /* restore \n (trailing \n not restored) */
+		if (q)		/* restore \n (trailing \n not restored) */
 			q[-1] = '\n';
 	}
 
-	/* Commands are executed here instead of pushing them onto the
-	 * input 'cause posix says the redirection and variable assignments
+	/*
+	 * Commands are executed here instead of pushing them onto the
+	 * input 'cause POSIX says the redirection and variable assignments
 	 * in
 	 *	X=y fc -e - 42 2> /dev/null
 	 * are to effect the repeated commands environment.
 	 */
 	/* XXX: source should not get trashed by this.. */
 	sold = source;
-	ret = command(cmd);
+	ret = command(cmd, 0);
 	source = sold;
-	return ret;
+	return (ret);
 }
 
 static int
@@ -306,7 +330,7 @@ hist_replace(char **hp, const char *pat, const char *rep, int globr)
 	char *line;
 
 	if (!pat)
-		line = str_save(*hp, ATEMP);
+		strdupx(line, *hp, ATEMP);
 	else {
 		char *s, *s1;
 		int pat_len = strlen(pat);
@@ -329,7 +353,7 @@ hist_replace(char **hp, const char *pat, const char *rep, int globr)
 		}
 		if (!any_subst) {
 			bi_errorf("substitution failed");
-			return 1;
+			return (1);
 		}
 		len = strlen(s) + 1;
 		XcheckN(xs, xp, len);
@@ -337,7 +361,7 @@ hist_replace(char **hp, const char *pat, const char *rep, int globr)
 		xp += len;
 		line = Xclose(xs, xp);
 	}
-	return hist_execute(line);
+	return (hist_execute(line));
 }
 
 /*
@@ -345,7 +369,7 @@ hist_replace(char **hp, const char *pat, const char *rep, int globr)
  * pattern is a number or string
  */
 static char **
-hist_get(const char *str, int approx, int allow_cur)
+hist_get(const char *str, bool approx, bool allow_cur)
 {
 	char **hp = NULL;
 	int n;
@@ -379,29 +403,29 @@ hist_get(const char *str, int approx, int allow_cur)
 		else
 			hp = &history[n];
 	}
-	return hp;
+	return (hp);
 }
 
 /* Return a pointer to the newest command in the history */
 char **
-hist_get_newest(int allow_cur)
+hist_get_newest(bool allow_cur)
 {
 	if (histptr < history || (!allow_cur && histptr == history)) {
 		bi_errorf("no history (yet)");
-		return NULL;
+		return (NULL);
 	}
 	return (allow_cur ? histptr : histptr - 1);
 }
 
-/* Return a pointer to the newest command in the history */
+/* Return a pointer to the oldest command in the history */
 static char **
 hist_get_oldest(void)
 {
 	if (histptr <= history) {
 		bi_errorf("no history (yet)");
-		return NULL;
+		return (NULL);
 	}
-	return history;
+	return (history);
 }
 
 /******************************/
@@ -414,7 +438,7 @@ histbackup(void)
 
 	if (histptr >= history && last_line != hist_source->line) {
 		hist_source->line--;
-		afree((void*)*histptr, APERM);
+		afree(*histptr, APERM);
 		histptr--;
 		last_line = hist_source->line;
 	}
@@ -426,7 +450,7 @@ histbackup(void)
 char **
 histpos(void)
 {
-	return current;
+	return (current);
 }
 
 int
@@ -436,10 +460,10 @@ histnum(int n)
 
 	if (n < 0 || n >= last) {
 		current = histptr;
-		return last;
+		return (last);
 	} else {
 		current = &history[n];
-		return n;
+		return (n);
 	}
 }
 
@@ -463,9 +487,9 @@ findhist(int start, int fwd, const char *str, int anchored)
 	for (; hp >= history && hp <= histptr; hp += incr)
 		if ((anchored && strncmp(*hp, str, len) == 0) ||
 		    (!anchored && strstr(*hp, str)))
-			return hp - history;
+			return (hp - history);
 
-	return -1;
+	return (-1);
 }
 
 int
@@ -477,15 +501,15 @@ findhistrel(const char *str)
 
 	getn(str, &rec);
 	if (rec == 0)
-		return -1;
+		return (-1);
 	if (rec > 0) {
 		if (rec > maxhist)
-			return -1;
-		return rec - 1;
+			return (-1);
+		return (rec - 1);
 	}
 	if (rec > maxhist)
-		return -1;
-	return start + rec + 1;
+		return (-1);
+	return (start + rec + 1);
 }
 
 /*
@@ -504,7 +528,7 @@ sethistsize(int n)
 			cursize = n;
 		}
 
-		history = (char **)aresize(history, n*sizeof(char *), APERM);
+		history = aresize(history, n * sizeof(char *), APERM);
 
 		histsize = n;
 		histptr = history + cursize;
@@ -533,7 +557,7 @@ sethistfile(const char *name)
 	 */
 	if (histfd) {
 		/* yes the file is open */
-		(void) close(histfd);
+		(void)close(histfd);
 		histfd = 0;
 		hsize = 0;
 		afree(hname, APERM);
@@ -555,7 +579,7 @@ init_histvec(void)
 {
 	if (history == (char **)NULL) {
 		histsize = HISTORYSIZE;
-		history = (char **)alloc(histsize * sizeof (char *), APERM);
+		history = alloc(histsize * sizeof(char *), APERM);
 		histptr = history - 1;
 	}
 }
@@ -569,29 +593,60 @@ init_histvec(void)
  *	It turns out that there is a lot of ghastly hackery here
  */
 
+#if !defined(MKSH_SMALL) && HAVE_PERSISTENT_HISTORY
+/* do not save command in history but possibly sync */
+bool
+histsync(void)
+{
+	bool changed = false;
+
+	if (histfd) {
+		int lno = hist_source->line;
+
+		hist_source->line++;
+		writehistfile(0, NULL);
+		hist_source->line--;
+
+		if (lno != hist_source->line)
+			changed = true;
+	}
+
+	return (changed);
+}
+#endif
 
 /*
  * save command in history
  */
 void
-histsave(int lno __unused, const char *cmd, int dowrite __unused)
+histsave(int *lnp, const char *cmd, bool dowrite MKSH_A_UNUSED, bool ignoredups)
 {
 	char **hp;
 	char *c, *cp;
 
-	c = str_save(cmd, APERM);
+	strdupx(c, cmd, APERM);
 	if ((cp = strchr(c, '\n')) != NULL)
 		*cp = '\0';
 
+	if (ignoredups && !strcmp(c, *histptr)
+#if !defined(MKSH_SMALL) && HAVE_PERSISTENT_HISTORY
+	    && !histsync()
+#endif
+	    ) {
+		afree(c, APERM);
+		return;
+	}
+	++*lnp;
+
 #if HAVE_PERSISTENT_HISTORY
 	if (histfd && dowrite)
-		writehistfile(lno, c);
+		writehistfile(*lnp, c);
 #endif
 
 	hp = histptr;
 
 	if (++hp >= history + histsize) { /* remove oldest command */
-		afree((void*)*history, APERM);
+		afree(*history, APERM);
 		for (hp = history; hp < history + histsize - 1; hp++)
 			hp[0] = hp[1];
 	}
@@ -613,8 +668,9 @@ histsave(int lno __unused, const char *cmd, int dowrite __unused)
 /*
  *	Open a history file
  *	Format is:
- *	Bytes 1, 2: HMAGIC - just to check that we are dealing with
- *		    the correct object
+ *	Bytes 1, 2:
+ *		HMAGIC - just to check that we are dealing with
+ *		the correct object
  *	Then follows a number of stored commands
  *	Each command is
  *	<command byte><command number(4 bytes)><bytes><null>
@@ -628,8 +684,7 @@ hist_init(Source *s)
 {
 #if HAVE_PERSISTENT_HISTORY
 	unsigned char *base;
-	int lines;
-	int fd;
+	int lines, fd, rv = 0;
 #endif
 
 	if (Flag(FTALKING) == 0)
@@ -642,7 +697,7 @@ hist_init(Source *s)
 #if HAVE_PERSISTENT_HISTORY
 	if ((hname = str_val(global("HISTFILE"))) == NULL)
 		return;
-	hname = str_save(hname, APERM);
+	strdupx(hname, hname, APERM);
 
  retry:
 	/* we have a file and are interactive */
@@ -653,9 +708,9 @@ hist_init(Source *s)
 	if (histfd != fd)
 		close(fd);
 
-	(void) flock(histfd, LOCK_EX);
+	(void)flock(histfd, LOCK_EX);
 
-	hsize = lseek(histfd, 0L, SEEK_END);
+	hsize = lseek(histfd, (off_t)0, SEEK_END);
 
 	if (hsize == 0) {
 		/* add magic */
@@ -668,7 +723,7 @@ hist_init(Source *s)
 		 * we have some data
 		 */
 		base = (void *)mmap(NULL, hsize, PROT_READ,
-		    MAP_FILE | MAP_PRIVATE, histfd, 0);
+		    MAP_FILE | MAP_PRIVATE, histfd, (off_t)0);
 		/*
 		 * check on its validity
 		 */
@@ -682,8 +737,6 @@ hist_init(Source *s)
 			goto retry;
 		}
 		if (hsize > 2) {
-			int rv = 0;
-
 			lines = hist_count_lines(base+2, hsize-2);
 			if (lines > histsize) {
 				/* we need to make the file smaller */
@@ -704,8 +757,8 @@ hist_init(Source *s)
 		histload(hist_source, base+2, hsize-2);
 		munmap((caddr_t)base, hsize);
 	}
-	(void) flock(histfd, LOCK_UN);
-	hsize = lseek(histfd, 0L, SEEK_END);
+	(void)flock(histfd, LOCK_UN);
+	hsize = lseek(histfd, (off_t)0, SEEK_END);
 #endif
 }
 
@@ -738,12 +791,14 @@ hist_count_lines(unsigned char *base, int bytes)
 		case sn4:
 			state = sline; break;
 		case sline:
-			if (*base == '\0')
-				lines++, state = shdr;
+			if (*base == '\0') {
+				lines++;
+				state = shdr;
+			}
 		}
 		base++;
 	}
-	return lines;
+	return (lines);
 }
 
 /*
@@ -752,42 +807,47 @@ hist_count_lines(unsigned char *base, int bytes)
 static int
 hist_shrink(unsigned char *oldbase, int oldbytes)
 {
-	int fd;
-	char	nfile[PATH_MAX];
+	int fd, rv = 0;
+	char *nfile = NULL;
 	struct	stat statb;
 	unsigned char *nbase = oldbase;
 	int nbytes = oldbytes;
 
 	nbase = hist_skip_back(nbase, &nbytes, histsize);
 	if (nbase == NULL)
-		return 1;
+		return (1);
 	if (nbase == oldbase)
-		return 0;
+		return (0);
 
 	/*
 	 *	create temp file
 	 */
-	shf_snprintf(nfile, sizeof(nfile), "%s.%d", hname, (int)procpid);
+	nfile = shf_smprintf("%s.%d", hname, (int)procpid);
 	if ((fd = open(nfile, O_CREAT | O_TRUNC | O_WRONLY, 0600)) < 0)
-		return 1;
+		goto errout;
 	if (fstat(histfd, &statb) >= 0 &&
 	    chown(nfile, statb.st_uid, statb.st_gid))
 		goto errout;
 
-	if (sprinkle(fd) || write(fd, nbase, nbytes) != nbytes) {
- errout:
-		close(fd);
-		unlink(nfile);
-		return 1;
-	}
+	if (sprinkle(fd) || write(fd, nbase, nbytes) != nbytes)
+		goto errout;
 	close(fd);
+	fd = -1;
 
 	/*
 	 *	rename
 	 */
-	if (rename(nfile, hname) < 0)
-		return 1;
-	return 0;
+	if (rename(nfile, hname) < 0) {
+ errout:
+		if (fd >= 0) {
+			close(fd);
+			if (nfile)
+				unlink(nfile);
+		}
+		rv = 1;
+	}
+	afree(nfile, ATEMP);
+	return (rv);
 }
 
 /*
@@ -801,9 +861,10 @@ hist_skip_back(unsigned char *base, int *bytes, int no)
 	unsigned char *ep;
 
 	for (ep = base + *bytes; --ep > base; ) {
-		/* this doesn't really work: the 4 byte line number that is
-		 * encoded after the COMMAND byte can itself contain the
-		 * COMMAND byte....
+		/*
+		 * this doesn't really work: the 4 byte line number that
+		 * is encoded after the COMMAND byte can itself contain
+		 * the COMMAND byte....
 		 */
 		for (; ep > base && *ep != COMMAND; ep--)
 			;
@@ -811,10 +872,10 @@ hist_skip_back(unsigned char *base, int *bytes, int no)
 			break;
 		if (++lines == no) {
 			*bytes = *bytes - ((char *)ep - (char *)base);
-			return ep;
+			return (ep);
 		}
 	}
-	return NULL;
+	return (NULL);
 }
 
 /*
@@ -824,8 +885,8 @@ static void
 histload(Source *s, unsigned char *base, int bytes)
 {
 	State state;
-	int	lno = 0;
-	unsigned char	*line = NULL;
+	int lno = 0;
+	unsigned char *line = NULL;
 
 	for (state = shdr; bytes-- > 0; base++) {
 		switch (state) {
@@ -857,8 +918,9 @@ histload(Source *s, unsigned char *base, int bytes)
 					/* a replacement ? */
 					histinsert(s, lno, (char *)line);
 				} else {
-					s->line = lno;
-					histsave(lno, (char *)line, 0);
+					s->line = lno--;
+					histsave(&lno, (char *)line, false,
+					    false);
 				}
 				state = shdr;
 			}
@@ -877,8 +939,8 @@ histinsert(Source *s, int lno, const char *line)
 	if (lno >= s->line - (histptr - history) && lno <= s->line) {
 		hp = &histptr[lno - s->line];
 		if (*hp)
-			afree((void*)*hp, APERM);
-		*hp = str_save(line, APERM);
+			afree(*hp, APERM);
+		strdupx(*hp, line, APERM);
 	}
 }
 
@@ -894,12 +956,12 @@ writehistfile(int lno, char *cmd)
 {
 	int	sizenow;
 	unsigned char	*base;
-	unsigned char	*new;
+	unsigned char	*news;
 	int	bytes;
 	unsigned char	hdr[5];
 
-	(void) flock(histfd, LOCK_EX);
-	sizenow = lseek(histfd, 0L, SEEK_END);
+	(void)flock(histfd, LOCK_EX);
+	sizenow = lseek(histfd, (off_t)0, SEEK_END);
 	if (sizenow != hsize) {
 		/*
 		 *	Things have changed
@@ -908,16 +970,16 @@ writehistfile(int lno, char *cmd)
 			/* someone has added some lines */
 			bytes = sizenow - hsize;
 			base = (void *)mmap(NULL, sizenow, PROT_READ,
-			    MAP_FILE | MAP_PRIVATE, histfd, 0);
+			    MAP_FILE | MAP_PRIVATE, histfd, (off_t)0);
 			if (base == (unsigned char *)MAP_FAILED)
 				goto bad;
-			new = base + hsize;
-			if (*new != COMMAND) {
+			news = base + hsize;
+			if (*news != COMMAND) {
 				munmap((caddr_t)base, sizenow);
 				goto bad;
 			}
 			hist_source->line--;
-			histload(hist_source, new, bytes);
+			histload(hist_source, news, bytes);
 			hist_source->line++;
 			lno = hist_source->line;
 			munmap((caddr_t)base, sizenow);
@@ -929,20 +991,22 @@ writehistfile(int lno, char *cmd)
 			goto bad;
 		}
 	}
-	/*
-	 *	we can write our bit now
-	 */
-	hdr[0] = COMMAND;
-	hdr[1] = (lno>>24)&0xff;
-	hdr[2] = (lno>>16)&0xff;
-	hdr[3] = (lno>>8)&0xff;
-	hdr[4] = lno&0xff;
-	bytes = strlen(cmd) + 1;
-	if ((write(histfd, hdr, 5) != 5) ||
-	    (write(histfd, cmd, bytes) != bytes))
-		goto bad;
-	hsize = lseek(histfd, 0L, SEEK_END);
-	(void) flock(histfd, LOCK_UN);
+	if (cmd) {
+		/*
+		 *	we can write our bit now
+		 */
+		hdr[0] = COMMAND;
+		hdr[1] = (lno>>24)&0xff;
+		hdr[2] = (lno>>16)&0xff;
+		hdr[3] = (lno>>8)&0xff;
+		hdr[4] = lno&0xff;
+		bytes = strlen(cmd) + 1;
+		if ((write(histfd, hdr, 5) != 5) ||
+		    (write(histfd, cmd, bytes) != bytes))
+			goto bad;
+		hsize = lseek(histfd, (off_t)0, SEEK_END);
+	}
+	(void)flock(histfd, LOCK_UN);
 	return;
  bad:
 	hist_finish();
@@ -951,8 +1015,8 @@ writehistfile(int lno, char *cmd)
 void
 hist_finish(void)
 {
-	(void) flock(histfd, LOCK_UN);
-	(void) close(histfd);
+	(void)flock(histfd, LOCK_UN);
+	(void)close(histfd);
 	histfd = 0;
 }
 
@@ -964,17 +1028,17 @@ sprinkle(int fd)
 {
 	static const unsigned char mag[] = { HMAGIC1, HMAGIC2 };
 
-	return(write(fd, mag, 2) != 2);
+	return (write(fd, mag, 2) != 2);
 }
 #endif
 
 #if !HAVE_SYS_SIGNAME
 static const struct mksh_sigpair {
-	int nr;
 	const char *const name;
+	int nr;
 } mksh_sigpairs[] = {
 #include "signames.inc"
-	{ 0, NULL }
+	{ NULL, 0 }
 };
 #endif
 
@@ -1007,7 +1071,8 @@ inittraps(void)
 
 				if (!strncasecmp(cs, "SIG", 3))
 					cs += 3;
-				sigtraps[i].name = s = str_save(cs, APERM);
+				strdupx(s, cs, APERM);
+				sigtraps[i].name = s;
 				while ((*s = ksh_toupper(*s)))
 					++s;
 			}
@@ -1025,7 +1090,7 @@ inittraps(void)
 	}
 	sigtraps[SIGEXIT_].name = "EXIT";	/* our name for signal 0 */
 
-	sigemptyset(&Sigact_ign.sa_mask);
+	(void)sigemptyset(&Sigact_ign.sa_mask);
 	Sigact_ign.sa_flags = 0; /* interruptible */
 	Sigact_ign.sa_handler = SIG_IGN;
 
@@ -1054,9 +1119,9 @@ alarm_init(void)
 
 /* ARGSUSED */
 static void
-alarm_catcher(int sig __unused)
+alarm_catcher(int sig MKSH_A_UNUSED)
 {
-	int errno_ = errno;
+	/* this runs inside interrupt context, with errno saved */
 
 	if (ksh_tmout_state == TMOUT_READING) {
 		int left = alarm(0);
@@ -1067,7 +1132,6 @@ alarm_catcher(int sig __unused)
 		} else
 			alarm(left);
 	}
-	errno = errno_;
 }
 
 Trap *
@@ -1115,7 +1179,8 @@ trapsig(int i)
 	errno = errno_;
 }
 
-/* called when we want to allow the user to ^C out of something - won't
+/*
+ * called when we want to allow the user to ^C out of something - won't
  * work if user has trapped SIGINT.
  */
 void
@@ -1125,7 +1190,8 @@ intrcheck(void)
 		runtraps(TF_DFL_INTR|TF_FATAL);
 }
 
-/* called after EINTR to check if a signal with normally causes process
+/*
+ * called after EINTR to check if a signal with normally causes process
  * termination has been received.
  */
 int
@@ -1138,11 +1204,12 @@ fatal_trap_check(void)
 	for (p = sigtraps, i = NSIG+1; --i >= 0; p++)
 		if (p->set && (p->flags & (TF_DFL_INTR|TF_FATAL)))
 			/* return value is used as an exit code */
-			return 128 + p->signal;
-	return 0;
+			return (128 + p->signal);
+	return (0);
 }
 
-/* Returns the signal number of any pending traps: ie, a signal which has
+/*
+ * Returns the signal number of any pending traps: ie, a signal which has
  * occurred for which a trap has been set or for which the TF_DFL_INTR flag
  * is set.
  */
@@ -1155,12 +1222,12 @@ trap_pending(void)
 	for (p = sigtraps, i = NSIG+1; --i >= 0; p++)
 		if (p->set && ((p->trap && p->trap[0]) ||
 		    ((p->flags & (TF_DFL_INTR|TF_FATAL)) && !p->trap)))
-			return p->signal;
-	return 0;
+			return (p->signal);
+	return (0);
 }
 
 /*
- * run any pending traps.  If intr is set, only run traps that
+ * run any pending traps. If intr is set, only run traps that
  * can interrupt commands.
  */
 void
@@ -1174,7 +1241,8 @@ runtraps(int flag)
 		warningf(false, "timed out waiting for input");
 		unwind(LEXIT);
 	} else
-		/* XXX: this means the alarm will have no effect if a trap
+		/*
+		 * XXX: this means the alarm will have no effect if a trap
 		 * is caught after the alarm() was started...not good.
 		 */
 		ksh_tmout_state = TMOUT_EXECUTING;
@@ -1220,10 +1288,11 @@ runtrap(Trap *p)
 		p->trap = NULL;
 	}
 	oexstat = exstat;
-	/* Note: trapstr is fully parsed before anything is executed, thus
+	/*
+	 * Note: trapstr is fully parsed before anything is executed, thus
 	 * no problem with afree(p->trap) in settrap() while still in use.
 	 */
-	command(trapstr);
+	command(trapstr, current_lineno);
 	exstat = oexstat;
 	if (i == SIGEXIT_ || i == SIGERR_) {
 		if (p->flags & TF_CHANGED)
@@ -1272,7 +1341,7 @@ settrap(Trap *p, const char *s)
 
 	if (p->trap)
 		afree(p->trap, APERM);
-	p->trap = str_save(s, APERM); /* handles s == 0 */
+	strdupx(p->trap, s, APERM); /* handles s == 0 */
 	p->flags |= TF_CHANGED;
 	f = !s ? SIG_DFL : s[0] ? trapsig : SIG_IGN;
 
@@ -1289,7 +1358,8 @@ settrap(Trap *p, const char *s)
 				p->flags |= TF_EXEC_DFL;
 		}
 
-		/* assumes handler already set to what shell wants it
+		/*
+		 * assumes handler already set to what shell wants it
 		 * (normally trapsig, but could be j_sigchld() or SIG_IGN)
 		 */
 		return;
@@ -1299,7 +1369,8 @@ settrap(Trap *p, const char *s)
 	setsig(p, f, SS_RESTORE_CURR|SS_USER);
 }
 
-/* Called by c_print() when writing to a co-process to ensure SIGPIPE won't
+/*
+ * Called by c_print() when writing to a co-process to ensure SIGPIPE won't
  * kill shell (unless user catches it and exits)
  */
 int
@@ -1316,7 +1387,7 @@ block_pipe(void)
 		setsig(p, SIG_IGN, SS_RESTORE_CURR);
 		restore_dfl = 1; /* restore to SIG_DFL */
 	}
-	return restore_dfl;
+	return (restore_dfl);
 }
 
 /* Called by c_print() to undo whatever block_pipe() did */
@@ -1327,7 +1398,8 @@ restore_pipe(int restore_dfl)
 		setsig(&sigtraps[SIGPIPE], SIG_DFL, SS_RESTORE_CURR);
 }
 
-/* Set action for a signal.  Action may not be set if original
+/*
+ * Set action for a signal. Action may not be set if original
  * action was SIG_IGN, depending on the value of flags and FTALKING.
  */
 int
@@ -1336,9 +1408,10 @@ setsig(Trap *p, sig_t f, int flags)
 	struct sigaction sigact;
 
 	if (p->signal == SIGEXIT_ || p->signal == SIGERR_)
-		return 1;
+		return (1);
 
-	/* First time setting this signal?  If so, get and note the current
+	/*
+	 * First time setting this signal? If so, get and note the current
 	 * setting.
 	 */
 	if (!(p->flags & (TF_ORIG_IGN|TF_ORIG_DFL))) {
@@ -1348,19 +1421,22 @@ setsig(Trap *p, sig_t f, int flags)
 		p->cursig = SIG_IGN;
 	}
 
-	/* Generally, an ignored signal stays ignored, except if
+	/*-
+	 * Generally, an ignored signal stays ignored, except if
 	 *	- the user of an interactive shell wants to change it
 	 *	- the shell wants for force a change
 	 */
 	if ((p->flags & TF_ORIG_IGN) && !(flags & SS_FORCE) &&
 	    (!(flags & SS_USER) || !Flag(FTALKING)))
-		return 0;
+		return (0);
 
 	setexecsig(p, flags & SS_RESTORE_MASK);
 
-	/* This is here 'cause there should be a way of clearing shtraps, but
-	 * don't know if this is a sane way of doing it.  At the moment,
-	 * all users of shtrap are lifetime users (SIGCHLD, SIGALRM, SIGWINCH).
+	/*
+	 * This is here 'cause there should be a way of clearing
+	 * shtraps, but don't know if this is a sane way of doing
+	 * it. At the moment, all users of shtrap are lifetime
+	 * users (SIGALRM, SIGCHLD, SIGWINCH).
 	 */
 	if (!(flags & SS_USER))
 		p->shtrap = (sig_t)NULL;
@@ -1371,13 +1447,13 @@ setsig(Trap *p, sig_t f, int flags)
 
 	if (p->cursig != f) {
 		p->cursig = f;
-		sigemptyset(&sigact.sa_mask);
+		(void)sigemptyset(&sigact.sa_mask);
 		sigact.sa_flags = 0 /* interruptible */;
 		sigact.sa_handler = f;
 		sigaction(p->signal, &sigact, NULL);
 	}
 
-	return 1;
+	return (1);
 }
 
 /* control what signal is set to before an exec() */
