@@ -1,8 +1,35 @@
-/*	$OpenBSD: jobs.c,v 1.36 2007/09/06 19:57:47 otto Exp $	*/
+/*	$OpenBSD: jobs.c,v 1.38 2009/12/12 04:28:44 deraadt Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/jobs.c,v 1.36 2008/05/17 19:03:25 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/jobs.c,v 1.69 2010/07/04 17:33:54 tg Exp $");
+
+#if HAVE_KILLPG
+#define mksh_killpg		killpg
+#else
+/* cross fingers and hope kill is killpg-endowed */
+#define mksh_killpg(p,s)	kill(-(p), (s))
+#endif
 
 /* Order important! */
 #define PRUNNING	0
@@ -13,9 +40,9 @@ __RCSID("$MirOS: src/bin/mksh/jobs.c,v 1.36 2008/05/17 19:03:25 tg Exp $");
 typedef struct proc	Proc;
 struct proc {
 	Proc *next;		/* next process in pipeline (if any) */
+	pid_t pid;		/* process id */
 	int state;
 	int status;		/* wait status */
-	pid_t pid;		/* process id */
 	char command[48];	/* process command string */
 };
 
@@ -47,20 +74,22 @@ struct proc {
 typedef struct job Job;
 struct job {
 	Job *next;		/* next job in list */
+	Proc *proc_list;	/* process list */
+	Proc *last_proc;	/* last process in list */
+	struct timeval systime;	/* system time used by job */
+	struct timeval usrtime;	/* user time used by job */
+	pid_t pgrp;		/* process group of job */
+	pid_t ppid;		/* pid of process that forked job */
 	int job;		/* job number: %n */
 	int flags;		/* see JF_* */
 	volatile int state;	/* job state */
 	int status;		/* exit status of last process */
-	pid_t pgrp;		/* process group of job */
-	pid_t ppid;		/* pid of process that forked job */
 	int32_t	age;		/* number of jobs started */
-	struct timeval systime;	/* system time used by job */
-	struct timeval usrtime;	/* user time used by job */
-	Proc *proc_list;	/* process list */
-	Proc *last_proc;	/* last process in list */
 	Coproc_id coproc_id;	/* 0 or id of coprocess output pipe */
+#ifndef MKSH_UNEMPLOYED
 	struct termios ttystate;/* saved tty state for stopped jobs */
 	pid_t saved_ttypgrp;	/* saved tty process group for stopped jobs */
+#endif
 };
 
 /* Flags for j_waitj() */
@@ -92,16 +121,18 @@ static int nzombie;		/* # of zombies owned by this process */
 static int32_t njobs;		/* # of jobs started */
 
 #ifndef CHILD_MAX
-#define CHILD_MAX	_POSIX_CHILD_MAX
+#define CHILD_MAX	25
 #endif
 
 /* held_sigchld is set if sigchld occurs before a job is completely started */
 static volatile sig_atomic_t held_sigchld;
 
+#ifndef MKSH_UNEMPLOYED
 static struct shf	*shl_j;
 static bool		ttypgrp_ok;	/* set if can use tty pgrps */
 static pid_t		restore_ttypgrp = -1;
 static int const	tt_sigs[] = { SIGTSTP, SIGTTIN, SIGTTOU };
+#endif
 
 static void		j_set_async(Job *);
 static void		j_startjob(Job *);
@@ -118,21 +149,29 @@ static int		kill_job(Job *, int);
 
 /* initialise job control */
 void
-j_init(int mflagset)
+j_init(void)
 {
-	sigemptyset(&sm_default);
+#ifndef MKSH_UNEMPLOYED
+	bool mflagset = Flag(FMONITOR) != 127;
+
+	Flag(FMONITOR) = 0;
+#endif
+
+	(void)sigemptyset(&sm_default);
 	sigprocmask(SIG_SETMASK, &sm_default, NULL);
 
-	sigemptyset(&sm_sigchld);
-	sigaddset(&sm_sigchld, SIGCHLD);
+	(void)sigemptyset(&sm_sigchld);
+	(void)sigaddset(&sm_sigchld, SIGCHLD);
 
 	setsig(&sigtraps[SIGCHLD], j_sigchld,
 	    SS_RESTORE_ORIG|SS_FORCE|SS_SHTRAP);
 
+#ifndef MKSH_UNEMPLOYED
 	if (!mflagset && Flag(FTALKING))
 		Flag(FMONITOR) = 1;
 
-	/* shl_j is used to do asynchronous notification (used in
+	/*
+	 * shl_j is used to do asynchronous notification (used in
 	 * an interrupt handler, so need a distinct shf)
 	 */
 	shl_j = shf_fdopen(2, SHF_WR, NULL);
@@ -140,7 +179,8 @@ j_init(int mflagset)
 	if (Flag(FMONITOR) || Flag(FTALKING)) {
 		int i;
 
-		/* the TF_SHELL_USES test is a kludge that lets us know if
+		/*
+		 * the TF_SHELL_USES test is a kludge that lets us know if
 		 * if the signals have been changed by the shell.
 		 */
 		for (i = NELEM(tt_sigs); --i >= 0; ) {
@@ -154,8 +194,10 @@ j_init(int mflagset)
 	/* j_change() calls tty_init() */
 	if (Flag(FMONITOR))
 		j_change();
-	else if (Flag(FTALKING))
-		tty_init(true);
+	else
+#endif
+	  if (Flag(FTALKING))
+		tty_init(true, true);
 }
 
 /* job cleanup before shell exit */
@@ -176,21 +218,25 @@ j_exit(void)
 			if (j->pgrp == 0)
 				kill_job(j, SIGHUP);
 			else
-				killpg(j->pgrp, SIGHUP);
+				mksh_killpg(j->pgrp, SIGHUP);
+#ifndef MKSH_UNEMPLOYED
 			if (j->state == PSTOPPED) {
 				if (j->pgrp == 0)
 					kill_job(j, SIGCONT);
 				else
-					killpg(j->pgrp, SIGCONT);
+					mksh_killpg(j->pgrp, SIGCONT);
 			}
+#endif
 		}
 	}
 	if (killed)
 		sleep(1);
 	j_notify();
 
+#ifndef MKSH_UNEMPLOYED
 	if (kshpid == procpid && restore_ttypgrp >= 0) {
-		/* Need to restore the tty pgrp to what it was when the
+		/*
+		 * Need to restore the tty pgrp to what it was when the
 		 * shell started up, so that the process that started us
 		 * will be able to access the tty when we are done.
 		 * Also need to restore our process group in case we are
@@ -204,8 +250,10 @@ j_exit(void)
 		Flag(FMONITOR) = 0;
 		j_change();
 	}
+#endif
 }
 
+#ifndef MKSH_UNEMPLOYED
 /* turn job control on or off according to Flag(FMONITOR) */
 void
 j_change(void)
@@ -217,7 +265,7 @@ j_change(void)
 
 		/* Don't call tcgetattr() 'til we own the tty process group */
 		if (use_tty)
-			tty_init(false);
+			tty_init(false, true);
 
 		/* no controlling tty, no SIGT* */
 		if ((ttypgrp_ok = use_tty && tty_fd >= 0 && tty_devtty)) {
@@ -282,148 +330,167 @@ j_change(void)
 			tty_close();
 	}
 }
+#endif
 
 /* execute tree in child subprocess */
 int
-exchild(struct op *t, int flags, /* used if XPCLOSE or XCCLOSE */ int close_fd)
+exchild(struct op *t, int flags,
+    volatile int *xerrok,
+    /* used if XPCLOSE or XCCLOSE */ int close_fd)
 {
-	static Proc	*last_proc;	/* for pipelines */
+	static Proc *last_proc;		/* for pipelines */
 
-	int		i;
-	sigset_t	omask;
-	Proc		*p;
-	Job		*j;
-	int		rv = 0;
-	int		forksleep;
-	int		ischild;
+	int rv = 0, forksleep;
+	sigset_t omask;
+	struct {
+		Proc *p;
+		Job *j;
+		pid_t cldpid;
+	} pi;
 
 	if (flags & XEXEC)
-		/* Clear XFORK|XPCLOSE|XCCLOSE|XCOPROC|XPIPEO|XPIPEI|XXCOM|XBGND
+		/*
+		 * Clear XFORK|XPCLOSE|XCCLOSE|XCOPROC|XPIPEO|XPIPEI|XXCOM|XBGND
 		 * (also done in another execute() below)
 		 */
-		return execute(t, flags & (XEXEC | XERROK));
+		return (execute(t, flags & (XEXEC | XERROK), xerrok));
 
 	/* no SIGCHLDs while messing with job and process lists */
 	sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
 
-	p = new_proc();
-	p->next = NULL;
-	p->state = PRUNNING;
-	p->status = 0;
-	p->pid = 0;
+	pi.p = new_proc();
+	pi.p->next = NULL;
+	pi.p->state = PRUNNING;
+	pi.p->status = 0;
+	pi.p->pid = 0;
 
 	/* link process into jobs list */
-	if (flags & XPIPEI) {	/* continuing with a pipe */
+	if (flags & XPIPEI) {
+		/* continuing with a pipe */
 		if (!last_job)
 			internal_errorf(
 			    "exchild: XPIPEI and no last_job - pid %d",
 			    (int)procpid);
-		j = last_job;
-		last_proc->next = p;
-		last_proc = p;
+		pi.j = last_job;
+		if (last_proc)
+			last_proc->next = pi.p;
+		last_proc = pi.p;
 	} else {
-		j = new_job(); /* fills in j->job */
-		/* we don't consider XXCOMs foreground since they don't get
+		pi.j = new_job(); /* fills in pi.j->job */
+		/*
+		 * we don't consider XXCOMs foreground since they don't get
 		 * tty process group and we don't save or restore tty modes.
 		 */
-		j->flags = (flags & XXCOM) ? JF_XXCOM :
+		pi.j->flags = (flags & XXCOM) ? JF_XXCOM :
 		    ((flags & XBGND) ? 0 : (JF_FG|JF_USETTYMODE));
-		timerclear(&j->usrtime);
-		timerclear(&j->systime);
-		j->state = PRUNNING;
-		j->pgrp = 0;
-		j->ppid = procpid;
-		j->age = ++njobs;
-		j->proc_list = p;
-		j->coproc_id = 0;
-		last_job = j;
-		last_proc = p;
-		put_job(j, PJ_PAST_STOPPED);
+		timerclear(&pi.j->usrtime);
+		timerclear(&pi.j->systime);
+		pi.j->state = PRUNNING;
+		pi.j->pgrp = 0;
+		pi.j->ppid = procpid;
+		pi.j->age = ++njobs;
+		pi.j->proc_list = pi.p;
+		pi.j->coproc_id = 0;
+		last_job = pi.j;
+		last_proc = pi.p;
+		put_job(pi.j, PJ_PAST_STOPPED);
 	}
 
-	snptreef(p->command, sizeof(p->command), "%T", t);
+	snptreef(pi.p->command, sizeof(pi.p->command), "%T", t);
 
 	/* create child process */
 	forksleep = 1;
-	while ((i = fork()) < 0 && errno == EAGAIN && forksleep < 32) {
+	while ((pi.cldpid = fork()) < 0 && errno == EAGAIN && forksleep < 32) {
 		if (intrsig)	 /* allow user to ^C out... */
 			break;
 		sleep(forksleep);
 		forksleep <<= 1;
 	}
-	if (i < 0) {
-		kill_job(j, SIGKILL);
-		remove_job(j, "fork failed");
+	if (pi.cldpid < 0) {
+		kill_job(pi.j, SIGKILL);
+		remove_job(pi.j, "fork failed");
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		errorf("cannot fork - try again");
 	}
-	ischild = i == 0;
-	if (ischild)
-		p->pid = procpid = getpid();
-	else
-		p->pid = i;
+	pi.p->pid = pi.cldpid ? pi.cldpid : (procpid = getpid());
 
-	/* Ensure next child gets a (slightly) different $RANDOM sequence */
-	change_random(((unsigned long)p->pid << 1) | (ischild ? 1 : 0));
+	/*
+	 * ensure next child gets a (slightly) different $RANDOM sequence
+	 * from its parent process and other child processes
+	 */
+	change_random(&pi, sizeof(pi));
 
+#ifndef MKSH_UNEMPLOYED
 	/* job control set up */
 	if (Flag(FMONITOR) && !(flags&XXCOM)) {
 		int	dotty = 0;
-		if (j->pgrp == 0) {	/* First process */
-			j->pgrp = p->pid;
+		if (pi.j->pgrp == 0) {	/* First process */
+			pi.j->pgrp = pi.p->pid;
 			dotty = 1;
 		}
 
 		/* set pgrp in both parent and child to deal with race
 		 * condition
 		 */
-		setpgid(p->pid, j->pgrp);
+		setpgid(pi.p->pid, pi.j->pgrp);
 		if (ttypgrp_ok && dotty && !(flags & XBGND))
-			tcsetpgrp(tty_fd, j->pgrp);
+			tcsetpgrp(tty_fd, pi.j->pgrp);
 	}
+#endif
 
 	/* used to close pipe input fd */
-	if (close_fd >= 0 && (((flags & XPCLOSE) && !ischild) ||
-	    ((flags & XCCLOSE) && ischild)))
+	if (close_fd >= 0 && (((flags & XPCLOSE) && pi.cldpid) ||
+	    ((flags & XCCLOSE) && !pi.cldpid)))
 		close(close_fd);
-	if (ischild) {		/* child */
+	if (!pi.cldpid) {
+		/* child */
+
 		/* Do this before restoring signal */
 		if (flags & XCOPROC)
 			coproc_cleanup(false);
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		cleanup_parents_env();
+#ifndef MKSH_UNEMPLOYED
 		/* If FMONITOR or FTALKING is set, these signals are ignored,
 		 * if neither FMONITOR nor FTALKING are set, the signals have
 		 * their inherited values.
 		 */
 		if (Flag(FMONITOR) && !(flags & XXCOM)) {
-			for (i = NELEM(tt_sigs); --i >= 0; )
-				setsig(&sigtraps[tt_sigs[i]], SIG_DFL,
+			for (forksleep = NELEM(tt_sigs); --forksleep >= 0; )
+				setsig(&sigtraps[tt_sigs[forksleep]], SIG_DFL,
 				    SS_RESTORE_DFL|SS_FORCE);
 		}
+#endif
+#if HAVE_NICE
 		if (Flag(FBGNICE) && (flags & XBGND))
-			i = nice(4);
-		if ((flags & XBGND) && !Flag(FMONITOR)) {
+			(void)nice(4);
+#endif
+		if ((flags & XBGND)
+#ifndef MKSH_UNEMPLOYED
+		    && !Flag(FMONITOR)
+#endif
+		    ) {
 			setsig(&sigtraps[SIGINT], SIG_IGN,
 			    SS_RESTORE_IGN|SS_FORCE);
 			setsig(&sigtraps[SIGQUIT], SIG_IGN,
 			    SS_RESTORE_IGN|SS_FORCE);
-			if (!(flags & (XPIPEI | XCOPROC))) {
-				int fd = open("/dev/null", 0);
-				if (fd != 0) {
-					(void) ksh_dup2(fd, 0, true);
-					close(fd);
-				}
+			if ((!(flags & (XPIPEI | XCOPROC))) &&
+			    ((forksleep = open("/dev/null", 0)) > 0)) {
+				(void)ksh_dup2(forksleep, 0, true);
+				close(forksleep);
 			}
 		}
-		remove_job(j, "child");	/* in case of $(jobs) command */
+		remove_job(pi.j, "child");	/* in case of $(jobs) command */
 		nzombie = 0;
+#ifndef MKSH_UNEMPLOYED
 		ttypgrp_ok = false;
 		Flag(FMONITOR) = 0;
+#endif
 		Flag(FTALKING) = 0;
 		tty_close();
 		cleartraps();
-		execute(t, (flags & XERROK) | XEXEC); /* no return */
+		/* no return */
+		execute(t, (flags & XERROK) | XEXEC, NULL);
 #ifndef MKSH_SMALL
 		if (t->type == TPIPE)
 			unwind(LLEAVE);
@@ -437,29 +504,32 @@ exchild(struct op *t, int flags, /* used if XPCLOSE or XCCLOSE */ int close_fd)
 
 	/* shell (parent) stuff */
 	if (!(flags & XPIPEO)) {	/* last process in a job */
-		j_startjob(j);
+		j_startjob(pi.j);
 		if (flags & XCOPROC) {
-			j->coproc_id = coproc.id;
-			coproc.njobs++; /* n jobs using co-process output */
-			coproc.job = (void *) j; /* j using co-process input */
+			pi.j->coproc_id = coproc.id;
+			/* n jobs using co-process output */
+			coproc.njobs++;
+			/* j using co-process input */
+			coproc.job = (void *)pi.j;
 		}
 		if (flags & XBGND) {
-			j_set_async(j);
+			j_set_async(pi.j);
 			if (Flag(FTALKING)) {
-				shf_fprintf(shl_out, "[%d]", j->job);
-				for (p = j->proc_list; p; p = p->next)
+				shf_fprintf(shl_out, "[%d]", pi.j->job);
+				for (pi.p = pi.j->proc_list; pi.p;
+				    pi.p = pi.p->next)
 					shf_fprintf(shl_out, " %d",
-					    (int)p->pid);
+					    (int)pi.p->pid);
 				shf_putchar('\n', shl_out);
 				shf_flush(shl_out);
 			}
 		} else
-			rv = j_waitj(j, JW_NONE, "jw:last proc");
+			rv = j_waitj(pi.j, JW_NONE, "jw:last proc");
 	}
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
 
-	return rv;
+	return (rv);
 }
 
 /* start the last job: only used for $(command) jobs */
@@ -495,14 +565,14 @@ waitlast(void)
 		else
 			internal_warningf("waitlast: not started");
 		sigprocmask(SIG_SETMASK, &omask, NULL);
-		return 125; /* not so arbitrary, non-zero value */
+		return (125); /* not so arbitrary, non-zero value */
 	}
 
 	rv = j_waitj(j, JW_NONE, "jw:waitlast");
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
 
-	return rv;
+	return (rv);
 }
 
 /* wait for child, interruptable. */
@@ -520,32 +590,33 @@ waitfor(const char *cp, int *sigp)
 	*sigp = 0;
 
 	if (cp == NULL) {
-		/* wait for an unspecified job - always returns 0, so
+		/*
+		 * wait for an unspecified job - always returns 0, so
 		 * don't have to worry about exited/signaled jobs
 		 */
 		for (j = job_list; j; j = j->next)
-			/* at&t ksh will wait for stopped jobs - we don't */
+			/* AT&T ksh will wait for stopped jobs - we don't */
 			if (j->ppid == procpid && j->state == PRUNNING)
 				break;
 		if (!j) {
 			sigprocmask(SIG_SETMASK, &omask, NULL);
-			return -1;
+			return (-1);
 		}
 	} else if ((j = j_lookup(cp, &ecode))) {
 		/* don't report normal job completion */
 		flags &= ~JW_ASYNCNOTIFY;
 		if (j->ppid != procpid) {
 			sigprocmask(SIG_SETMASK, &omask, NULL);
-			return -1;
+			return (-1);
 		}
 	} else {
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		if (ecode != JL_NOSUCH)
 			bi_errorf("%s: %s", cp, lookup_msgs[ecode]);
-		return -1;
+		return (-1);
 	}
 
-	/* at&t ksh will wait for stopped jobs - we don't */
+	/* AT&T ksh will wait for stopped jobs - we don't */
 	rv = j_waitj(j, flags, "jw:waitfor");
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
@@ -553,7 +624,7 @@ waitfor(const char *cp, int *sigp)
 	if (rv < 0) /* we were interrupted */
 		*sigp = 128 + -rv;
 
-	return rv;
+	return (rv);
 }
 
 /* kill (built-in) a job */
@@ -570,7 +641,7 @@ j_kill(const char *cp, int sig)
 	if ((j = j_lookup(cp, &ecode)) == NULL) {
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		bi_errorf("%s: %s", cp, lookup_msgs[ecode]);
-		return 1;
+		return (1);
 	}
 
 	if (j->pgrp == 0) {	/* started when !Flag(FMONITOR) */
@@ -579,9 +650,11 @@ j_kill(const char *cp, int sig)
 			rv = 1;
 		}
 	} else {
+#ifndef MKSH_UNEMPLOYED
 		if (j->state == PSTOPPED && (sig == SIGTERM || sig == SIGHUP))
-			(void) killpg(j->pgrp, SIGCONT);
-		if (killpg(j->pgrp, sig) < 0) {
+			mksh_killpg(j->pgrp, SIGCONT);
+#endif
+		if (mksh_killpg(j->pgrp, sig) < 0) {
 			bi_errorf("%s: %s", cp, strerror(errno));
 			rv = 1;
 		}
@@ -589,9 +662,10 @@ j_kill(const char *cp, int sig)
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
 
-	return rv;
+	return (rv);
 }
 
+#ifndef MKSH_UNEMPLOYED
 /* fg and bg built-ins: called only if Flag(FMONITOR) set */
 int
 j_resume(const char *cp, int bg)
@@ -608,13 +682,13 @@ j_resume(const char *cp, int bg)
 	if ((j = j_lookup(cp, &ecode)) == NULL) {
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		bi_errorf("%s: %s", cp, lookup_msgs[ecode]);
-		return 1;
+		return (1);
 	}
 
 	if (j->pgrp == 0) {
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		bi_errorf("job not job-controlled");
-		return 1;
+		return (1);
 	}
 
 	if (bg)
@@ -655,10 +729,10 @@ j_resume(const char *cp, int bg)
 				    NULL);
 				bi_errorf("1st tcsetpgrp(%d, %d) failed: %s",
 				    tty_fd,
-				    (int) ((j->flags & JF_SAVEDTTYPGRP) ?
+				    (int)((j->flags & JF_SAVEDTTYPGRP) ?
 				    j->saved_ttypgrp : j->pgrp),
 				    strerror(rv));
-				return 1;
+				return (1);
 			}
 		}
 		j->flags |= JF_FG;
@@ -667,7 +741,7 @@ j_resume(const char *cp, int bg)
 			async_job = NULL;
 	}
 
-	if (j->state == PRUNNING && killpg(j->pgrp, SIGCONT) < 0) {
+	if (j->state == PRUNNING && mksh_killpg(j->pgrp, SIGCONT) < 0) {
 		int err = errno;
 
 		if (!bg) {
@@ -682,7 +756,7 @@ j_resume(const char *cp, int bg)
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 		bi_errorf("cannot continue job %s: %s",
 		    cp, strerror(err));
-		return 1;
+		return (1);
 	}
 	if (!bg) {
 		if (ttypgrp_ok) {
@@ -691,8 +765,9 @@ j_resume(const char *cp, int bg)
 		rv = j_waitj(j, JW_NONE, "jw:resume");
 	}
 	sigprocmask(SIG_SETMASK, &omask, NULL);
-	return rv;
+	return (rv);
 }
+#endif
 
 /* are there any running or stopped jobs ? */
 int
@@ -702,8 +777,10 @@ j_stopped_running(void)
 	int	which = 0;
 
 	for (j = job_list; j != NULL; j = j->next) {
+#ifndef MKSH_UNEMPLOYED
 		if (j->ppid == procpid && j->state == PSTOPPED)
 			which |= 1;
+#endif
 		if (Flag(FLOGIN) && !Flag(FNOHUP) && procpid == kshpid &&
 		    j->ppid == procpid && j->state == PRUNNING)
 			which |= 2;
@@ -713,10 +790,10 @@ j_stopped_running(void)
 		    which & 1 ? "stopped" : "",
 		    which == 3 ? " and " : "",
 		    which & 2 ? "running" : "");
-		return 1;
+		return (1);
 	}
 
-	return 0;
+	return (0);
 }
 
 int
@@ -731,7 +808,7 @@ j_njobs(void)
 		nj++;
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
-	return nj;
+	return (nj);
 }
 
 
@@ -757,7 +834,7 @@ j_jobs(const char *cp, int slp,
 		if ((j = j_lookup(cp, &ecode)) == NULL) {
 			sigprocmask(SIG_SETMASK, &omask, NULL);
 			bi_errorf("%s: %s", cp, lookup_msgs[ecode]);
-			return 1;
+			return (1);
 		}
 	} else
 		j = job_list;
@@ -779,7 +856,7 @@ j_jobs(const char *cp, int slp,
 			remove_job(j, "jobs");
 	}
 	sigprocmask(SIG_SETMASK, &omask, NULL);
-	return 0;
+	return (0);
 }
 
 /* list jobs for top-level notification */
@@ -791,8 +868,10 @@ j_notify(void)
 
 	sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
 	for (j = job_list; j; j = j->next) {
+#ifndef MKSH_UNEMPLOYED
 		if (Flag(FMONITOR) && (j->flags & JF_CHANGED))
 			j_print(j, JP_MEDIUM, shl_out);
+#endif
 		/* Remove job after doing reports so there aren't
 		 * multiple +/- jobs.
 		 */
@@ -821,10 +900,11 @@ j_async(void)
 
 	sigprocmask(SIG_SETMASK, &omask, NULL);
 
-	return async_pid;
+	return (async_pid);
 }
 
-/* Make j the last async process
+/*
+ * Make j the last async process
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -860,7 +940,8 @@ j_set_async(Job *j)
 	}
 }
 
-/* Start a job: set STARTED, check for held signals and set j->last_proc
+/*
+ * Start a job: set STARTED, check for held signals and set j->last_proc
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -900,7 +981,9 @@ j_waitj(Job *j,
 	if (flags & JW_ASYNCNOTIFY)
 		j->flags |= JF_W_ASYNCNOTIFY;
 
+#ifndef MKSH_UNEMPLOYED
 	if (!Flag(FMONITOR))
+#endif
 		flags |= JW_STOPPEDWAIT;
 
 	while (j->state == PRUNNING ||
@@ -914,19 +997,18 @@ j_waitj(Job *j,
 		}
 		if ((flags & JW_INTERRUPT) && (rv = trap_pending())) {
 			j->flags &= ~(JF_WAITING|JF_W_ASYNCNOTIFY);
-			return -rv;
+			return (-rv);
 		}
 	}
 	j->flags &= ~(JF_WAITING|JF_W_ASYNCNOTIFY);
 
 	if (j->flags & JF_FG) {
-		int	status;
-
 		j->flags &= ~JF_FG;
+#ifndef MKSH_UNEMPLOYED
 		if (Flag(FMONITOR) && ttypgrp_ok && j->pgrp) {
 			/*
 			 * Save the tty's current pgrp so it can be restored
-			 * when the job is foregrounded.  This is to
+			 * when the job is foregrounded. This is to
 			 * deal with things like the GNU su which does
 			 * a fork/exec instead of an exec (the fork means
 			 * the execed shell gets a different pid from its
@@ -947,9 +1029,11 @@ j_waitj(Job *j,
 				tcgetattr(tty_fd, &j->ttystate);
 			}
 		}
+#endif
 		if (tty_fd >= 0) {
-			/* Only restore tty settings if job was originally
-			 * started in the foreground.  Problems can be
+			/*
+			 * Only restore tty settings if job was originally
+			 * started in the foreground. Problems can be
 			 * caused by things like 'more foobar &' which will
 			 * typically get and save the shell's vi/emacs tty
 			 * settings before setting up the tty for itself;
@@ -961,8 +1045,9 @@ j_waitj(Job *j,
 				tcgetattr(tty_fd, &tty_state);
 			} else {
 				tcsetattr(tty_fd, TCSADRAIN, &tty_state);
-				/* Don't use tty mode if job is stopped and
-				 * later restarted and exits.  Consider
+				/*-
+				 * Don't use tty mode if job is stopped and
+				 * later restarted and exits. Consider
 				 * the sequence:
 				 *	vi foo (stopped)
 				 *	...
@@ -976,50 +1061,66 @@ j_waitj(Job *j,
 					j->flags &= ~JF_USETTYMODE;
 			}
 		}
-		/* If it looks like user hit ^C to kill a job, pretend we got
-		 * one too to break out of for loops, etc.  (at&t ksh does this
+#ifndef MKSH_UNEMPLOYED
+		/*
+		 * If it looks like user hit ^C to kill a job, pretend we got
+		 * one too to break out of for loops, etc. (AT&T ksh does this
 		 * even when not monitoring, but this doesn't make sense since
 		 * a tty generated ^C goes to the whole process group)
 		 */
-		status = j->last_proc->status;
-		if (Flag(FMONITOR) && j->state == PSIGNALLED &&
-		    WIFSIGNALED(status) &&
-		    (sigtraps[WTERMSIG(status)].flags & TF_TTY_INTR))
-			trapsig(WTERMSIG(status));
+		{
+			int status;
+
+			status = j->last_proc->status;
+			if (Flag(FMONITOR) && j->state == PSIGNALLED &&
+			    WIFSIGNALED(status) &&
+			    (sigtraps[WTERMSIG(status)].flags & TF_TTY_INTR))
+				trapsig(WTERMSIG(status));
+		}
+#endif
 	}
 
 	j_usrtime = j->usrtime;
 	j_systime = j->systime;
 	rv = j->status;
 
-	if (!(flags & JW_ASYNCNOTIFY) &&
-	    (!Flag(FMONITOR) || j->state != PSTOPPED)) {
+	if (!(flags & JW_ASYNCNOTIFY)
+#ifndef MKSH_UNEMPLOYED
+	    && (!Flag(FMONITOR) || j->state != PSTOPPED)
+#endif
+	    ) {
 		j_print(j, JP_SHORT, shl_out);
 		shf_flush(shl_out);
 	}
-	if (j->state != PSTOPPED &&
-	    (!Flag(FMONITOR) || !(flags & JW_ASYNCNOTIFY)))
+	if (j->state != PSTOPPED
+#ifndef MKSH_UNEMPLOYED
+	    && (!Flag(FMONITOR) || !(flags & JW_ASYNCNOTIFY))
+#endif
+	    )
 		remove_job(j, where);
 
-	return rv;
+	return (rv);
 }
 
-/* SIGCHLD handler to reap children and update job states
+/*
+ * SIGCHLD handler to reap children and update job states
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
 /* ARGSUSED */
 static void
-j_sigchld(int sig __unused)
+j_sigchld(int sig MKSH_A_UNUSED)
 {
-	int		errno_ = errno;
-	Job		*j;
-	Proc		*p = NULL;
-	int		pid;
-	int		status;
-	struct rusage	ru0, ru1;
+	/* this runs inside interrupt context, with errno saved */
 
-	/* Don't wait for any processes if a job is partially started.
+	Job *j;
+	Proc *p = NULL;
+	pid_t pid;
+	int status;
+	struct rusage ru0, ru1;
+
+	/*
+	 * Don't wait for any processes if a job is partially started.
 	 * This is so we don't do away with the process group leader
 	 * before all the processes in a pipe line are started (so the
 	 * setpgid() won't fail)
@@ -1034,8 +1135,12 @@ j_sigchld(int sig __unused)
 	do {
 		pid = waitpid(-1, &status, (WNOHANG|WUNTRACED));
 
-		if (pid <= 0)	/* return if would block (0) ... */
-			break;	/* ... or no children or interrupted (-1) */
+		/*
+		 * return if this would block (0) or no children
+		 * or interrupted (-1)
+		 */
+		if (pid <= 0)
+			return;
 
 		getrusage(RUSAGE_CHILDREN, &ru1);
 
@@ -1060,23 +1165,23 @@ j_sigchld(int sig __unused)
 		timersub(&j->systime, &ru0.ru_stime, &j->systime);
 		ru0 = ru1;
 		p->status = status;
+#ifndef MKSH_UNEMPLOYED
 		if (WIFSTOPPED(status))
 			p->state = PSTOPPED;
-		else if (WIFSIGNALED(status))
+		else
+#endif
+		  if (WIFSIGNALED(status))
 			p->state = PSIGNALLED;
 		else
 			p->state = PEXITED;
 
 		check_job(j);	/* check to see if entire job is done */
-	}
-	while (1);
-
-	errno = errno_;
+	} while (1);
 }
 
 /*
  * Called only when a process in j has exited/stopped (ie, called only
- * from j_sigchld()).  If no processes are running, the job status
+ * from j_sigchld()). If no processes are running, the job status
  * and state are updated, asynchronous job notification is done and,
  * if unneeded, the job is removed.
  *
@@ -1116,17 +1221,20 @@ check_job(Job *j)
 		break;
 	}
 
-	/* Note when co-process dies: can't be done in j_wait() nor
+	/*
+	 * Note when co-process dies: can't be done in j_wait() nor
 	 * remove_job() since neither may be called for non-interactive
 	 * shells.
 	 */
 	if (j->state == PEXITED || j->state == PSIGNALLED) {
-		/* No need to keep co-process input any more
+		/*
+		 * No need to keep co-process input any more
 		 * (at least, this is what ksh93d thinks)
 		 */
 		if (coproc.job == j) {
 			coproc.job = NULL;
-			/* XXX would be nice to get the closes out of here
+			/*
+			 * XXX would be nice to get the closes out of here
 			 * so they aren't done in the signal handler.
 			 * Would mean a check in coproc_getfd() to
 			 * do "if job == 0 && write >= 0, close write".
@@ -1140,8 +1248,10 @@ check_job(Job *j)
 	}
 
 	j->flags |= JF_CHANGED;
+#ifndef MKSH_UNEMPLOYED
 	if (Flag(FMONITOR) && !(j->flags & JF_XXCOM)) {
-		/* Only put stopped jobs at the front to avoid confusing
+		/*
+		 * Only put stopped jobs at the front to avoid confusing
 		 * the user (don't want finished jobs effecting %+ or %-)
 		 */
 		if (j->state == PSTOPPED)
@@ -1158,7 +1268,8 @@ check_job(Job *j)
 						fd = ep->savefd[2];
 				shf_reopen(fd, SHF_WR, shl_j);
 			}
-			/* Can't call j_notify() as it removes jobs.  The job
+			/*
+			 * Can't call j_notify() as it removes jobs. The job
 			 * must stay in the job list as j_waitj() may be
 			 * running with this job.
 			 */
@@ -1168,7 +1279,12 @@ check_job(Job *j)
 				remove_job(j, "notify");
 		}
 	}
-	if (!Flag(FMONITOR) && !(j->flags & (JF_WAITING|JF_FG)) &&
+#endif
+	if (
+#ifndef MKSH_UNEMPLOYED
+	    !Flag(FMONITOR) &&
+#endif
+	    !(j->flags & (JF_WAITING|JF_FG)) &&
 	    j->state != PSTOPPED) {
 		if (j == async_job || (j->flags & JF_KNOWN)) {
 			j->flags |= JF_ZOMBIE;
@@ -1197,8 +1313,9 @@ j_print(Job *j, int how, struct shf *shf)
 	int	output = 0;
 
 	if (how == JP_PGRP) {
-		/* POSIX doesn't say what to do it there is no process
-		 * group leader (ie, !FMONITOR).  We arbitrarily return
+		/*
+		 * POSIX doesn't say what to do it there is no process
+		 * group leader (ie, !FMONITOR). We arbitrarily return
 		 * last pid (which is what $! returns).
 		 */
 		shf_fprintf(shf, "%d\n", (int)(j->pgrp ? j->pgrp :
@@ -1206,7 +1323,7 @@ j_print(Job *j, int how, struct shf *shf)
 		return;
 	}
 	j->flags &= ~JF_CHANGED;
-	filler = j->job > 10 ?  "\n       " : "\n      ";
+	filler = j->job > 10 ? "\n       " : "\n      ";
 	if (j == job_list)
 		jobchar = '+';
 	else if (j == job_list->next)
@@ -1216,17 +1333,17 @@ j_print(Job *j, int how, struct shf *shf)
 		coredumped = 0;
 		switch (p->state) {
 		case PRUNNING:
-			strlcpy(buf, "Running", sizeof buf);
+			memcpy(buf, "Running", 8);
 			break;
 		case PSTOPPED:
 			strlcpy(buf, sigtraps[WSTOPSIG(p->status)].mess,
-			    sizeof buf);
+			    sizeof(buf));
 			break;
 		case PEXITED:
 			if (how == JP_SHORT)
 				buf[0] = '\0';
 			else if (WEXITSTATUS(p->status) == 0)
-				strlcpy(buf, "Done", sizeof buf);
+				memcpy(buf, "Done", 5);
 			else
 				shf_snprintf(buf, sizeof(buf), "Done (%d)",
 				    WEXITSTATUS(p->status));
@@ -1236,8 +1353,9 @@ j_print(Job *j, int how, struct shf *shf)
 			if (WCOREDUMP(p->status))
 				coredumped = 1;
 #endif
-			/* kludge for not reporting 'normal termination signals'
-			 * (ie, SIGINT, SIGPIPE)
+			/*
+			 * kludge for not reporting 'normal termination
+			 * signals' (i.e. SIGINT, SIGPIPE)
 			 */
 			if (how == JP_SHORT && !coredumped &&
 			    (WTERMSIG(p->status) == SIGINT ||
@@ -1245,7 +1363,7 @@ j_print(Job *j, int how, struct shf *shf)
 				buf[0] = '\0';
 			} else
 				strlcpy(buf, sigtraps[WTERMSIG(p->status)].mess,
-				    sizeof buf);
+				    sizeof(buf));
 			break;
 		}
 
@@ -1290,7 +1408,8 @@ j_print(Job *j, int how, struct shf *shf)
 		shf_putc('\n', shf);
 }
 
-/* Convert % sequence to job
+/*
+ * Convert % sequence to job
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -1306,32 +1425,34 @@ j_lookup(const char *cp, int *ecodep)
 		/* Look for last_proc->pid (what $! returns) first... */
 		for (j = job_list; j != NULL; j = j->next)
 			if (j->last_proc && j->last_proc->pid == job)
-				return j;
-		/* ...then look for process group (this is non-POSIX,
-		 * but should not break anything */
+				return (j);
+		/*
+		 * ...then look for process group (this is non-POSIX,
+		 * but should not break anything
+		 */
 		for (j = job_list; j != NULL; j = j->next)
 			if (j->pgrp && j->pgrp == job)
-				return j;
+				return (j);
 		if (ecodep)
 			*ecodep = JL_NOSUCH;
-		return NULL;
+		return (NULL);
 	}
 	if (*cp != '%') {
 		if (ecodep)
 			*ecodep = JL_INVALID;
-		return NULL;
+		return (NULL);
 	}
 	switch (*++cp) {
 	case '\0': /* non-standard */
 	case '+':
 	case '%':
 		if (job_list != NULL)
-			return job_list;
+			return (job_list);
 		break;
 
 	case '-':
 		if (job_list != NULL && job_list->next)
-			return job_list->next;
+			return (job_list->next);
 		break;
 
 	case '0': case '1': case '2': case '3': case '4':
@@ -1339,7 +1460,7 @@ j_lookup(const char *cp, int *ecodep)
 		getn(cp, &job);
 		for (j = job_list; j != NULL; j = j->next)
 			if (j->job == job)
-				return j;
+				return (j);
 		break;
 
 	case '?':		/* %?string */
@@ -1350,12 +1471,12 @@ j_lookup(const char *cp, int *ecodep)
 					if (last_match) {
 						if (ecodep)
 							*ecodep = JL_AMBIG;
-						return NULL;
+						return (NULL);
 					}
 					last_match = j;
 				}
 		if (last_match)
-			return last_match;
+			return (last_match);
 		break;
 
 	default:		/* %string */
@@ -1366,23 +1487,24 @@ j_lookup(const char *cp, int *ecodep)
 				if (last_match) {
 					if (ecodep)
 						*ecodep = JL_AMBIG;
-					return NULL;
+					return (NULL);
 				}
 				last_match = j;
 			}
 		if (last_match)
-			return last_match;
+			return (last_match);
 		break;
 	}
 	if (ecodep)
 		*ecodep = JL_NOSUCH;
-	return NULL;
+	return (NULL);
 }
 
 static Job	*free_jobs;
 static Proc	*free_procs;
 
-/* allocate a new job and fill in the job number.
+/*
+ * allocate a new job and fill in the job number.
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -1396,7 +1518,7 @@ new_job(void)
 		newj = free_jobs;
 		free_jobs = free_jobs->next;
 	} else
-		newj = (Job *)alloc(sizeof (Job), APERM);
+		newj = alloc(sizeof(Job), APERM);
 
 	/* brute force method */
 	for (i = 1; ; i++) {
@@ -1407,10 +1529,11 @@ new_job(void)
 	}
 	newj->job = i;
 
-	return newj;
+	return (newj);
 }
 
-/* Allocate new process struct
+/*
+ * Allocate new process struct
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -1423,12 +1546,13 @@ new_proc(void)
 		p = free_procs;
 		free_procs = free_procs->next;
 	} else
-		p = (Proc *)alloc(sizeof (Proc), APERM);
+		p = alloc(sizeof(Proc), APERM);
 
-	return p;
+	return (p);
 }
 
-/* Take job out of job_list and put old structures into free list.
+/*
+ * Take job out of job_list and put old structures into free list.
  * Keeps nzombies, last_job and async_job up to date.
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
@@ -1468,7 +1592,8 @@ remove_job(Job *j, const char *where)
 		async_job = NULL;
 }
 
-/* put j in a particular location (taking it out job_list if it is there
+/*
+ * put j in a particular location (taking it out job_list if it is there
  * already)
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
@@ -1504,7 +1629,8 @@ put_job(Job *j, int where)
 	}
 }
 
-/* nuke a job (called when unable to start full job).
+/*
+ * nuke a job (called when unable to start full job).
  *
  * If jobs are compiled in then this routine expects sigchld to be blocked.
  */
@@ -1518,5 +1644,5 @@ kill_job(Job *j, int sig)
 		if (p->pid != 0)
 			if (kill(p->pid, sig) < 0)
 				rval = -1;
-	return rval;
+	return (rval);
 }
