@@ -1,8 +1,28 @@
 /*	$OpenBSD: lex.c,v 1.44 2008/07/03 17:52:08 otto Exp $	*/
 
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
+
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/lex.c,v 1.66 2008/07/12 16:56:39 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/lex.c,v 1.117 2010/07/21 11:31:15 tg Exp $");
 
 /*
  * states while lexing word
@@ -12,17 +32,19 @@ __RCSID("$MirOS: src/bin/mksh/lex.c,v 1.66 2008/07/12 16:56:39 tg Exp $");
 #define SLETPAREN	2	/* inside (( )), implicit quoting */
 #define SSQUOTE		3	/* inside '' */
 #define SDQUOTE		4	/* inside "" */
-#define SBRACE		5	/* inside ${} */
-#define SCSPAREN	6	/* inside $() */
-#define SBQUOTE		7	/* inside `` */
-#define SASPAREN	8	/* inside $(( )) */
-#define SHEREDELIM	9	/* parsing <<,<<- delimiter */
-#define SHEREDQUOTE	10	/* parsing " in <<,<<- delimiter */
-#define SPATTERN	11	/* parsing *(...|...) pattern (*+?@!) */
-#define STBRACE		12	/* parsing ${..[#%]..} */
-#define SLETARRAY	13	/* inside =( ), just copy */
-#define SADELIM		14	/* like SBASE, looking for delimiter */
-#define SHERESTRING	15	/* parsing <<< string */
+#define SEQUOTE		5	/* inside $'' */
+#define SBRACE		6	/* inside ${} */
+#define SQBRACE		7	/* inside "${}" */
+#define SCSPAREN	8	/* inside $() */
+#define SBQUOTE		9	/* inside `` */
+#define SASPAREN	10	/* inside $(( )) */
+#define SHEREDELIM	11	/* parsing <<,<<- delimiter */
+#define SHEREDQUOTE	12	/* parsing " in <<,<<- delimiter */
+#define SPATTERN	13	/* parsing *(...|...) pattern (*+?@!) */
+#define STBRACE		14	/* parsing ${...[#%]...} */
+#define SLETARRAY	15	/* inside =( ), just copy */
+#define SADELIM		16	/* like SBASE, looking for delimiter */
+#define SHERESTRING	17	/* parsing <<< string */
 
 /* Structure to keep track of the lexing state and the various pieces of info
  * needed for each particular state. */
@@ -56,11 +78,13 @@ struct lex_state {
 #define ls_sbquote ls_info.u_sbquote
 		} u_sbquote;
 
+#ifndef MKSH_SMALL
 		/* =(...) */
 		struct sletarray_info {
 			int nparen;	/* count open parentheses */
 #define ls_sletarray ls_info.u_sletarray
 		} u_sletarray;
+#endif
 
 		/* ADELIM */
 		struct sadelim_info {
@@ -74,59 +98,83 @@ struct lex_state {
 #define ls_sadelim ls_info.u_sadelim
 		} u_sadelim;
 
+		/* $'...' */
+		struct sequote_info {
+			bool got_NUL;	/* ignore rest of string */
+#define ls_sequote ls_info.u_sequote
+		} u_sequote;
+
 		Lex_state *base;	/* used to point to next state block */
 	} ls_info;
 };
 
-typedef struct State_info State_info;
-struct State_info {
+typedef struct {
 	Lex_state *base;
 	Lex_state *end;
-};
+} State_info;
 
-static void	readhere(struct ioword *);
-static int	getsc__(void);
-static void	getsc_line(Source *);
-static int	getsc_bn(void);
-static char	*get_brace_var(XString *, char *);
-static int	arraysub(char **);
+static void readhere(struct ioword *);
+static int getsc__(void);
+static void getsc_line(Source *);
+static int getsc_bn(void);
+static int s_get(void);
+static void s_put(int);
+static char *get_brace_var(XString *, char *);
+static int arraysub(char **);
 static const char *ungetsc(int);
-static void	gethere(bool);
+static void gethere(bool);
 static Lex_state *push_state_(State_info *, Lex_state *);
 static Lex_state *pop_state_(State_info *, Lex_state *);
 
-static int dopprompt(const char *, int, int);
+static int dopprompt(const char *, int, bool);
 
 static int backslash_skip;
 static int ignore_backslash_newline;
 
 /* optimised getsc_bn() */
-#define getsc()		(*source->str != '\0' && *source->str != '\\' \
+#define _getsc()	(*source->str != '\0' && *source->str != '\\' \
 			 && !backslash_skip && !(source->flags & SF_FIRST) \
 			 ? *source->str++ : getsc_bn())
 /* optimised getsc__() */
-#define	getsc_()	((*source->str != '\0') && !(source->flags & SF_FIRST) \
+#define	_getsc_()	((*source->str != '\0') && !(source->flags & SF_FIRST) \
 			 ? *source->str++ : getsc__())
 
 #ifdef MKSH_SMALL
-#define STATE_BSIZE	32
+static int getsc(void);
+static int getsc_(void);
+
+static int
+getsc(void)
+{
+	return (_getsc());
+}
+
+static int
+getsc_(void)
+{
+	return (_getsc_());
+}
 #else
-#define STATE_BSIZE	48
+/* !MKSH_SMALL: use them inline */
+#define getsc()		_getsc()
+#define getsc_()	_getsc_()
 #endif
 
-#define PUSH_STATE(s)	do { \
-			    if (++statep == state_info.end) \
-				statep = push_state_(&state_info, statep); \
-			    state = statep->ls_state = (s); \
-			} while (0)
+#define STATE_BSIZE	32
 
-#define POP_STATE()	do { \
-			    if (--statep == state_info.base) \
-				statep = pop_state_(&state_info, statep); \
-			    state = statep->ls_state; \
-			} while (0)
+#define PUSH_STATE(s)	do {					\
+	if (++statep == state_info.end)				\
+		statep = push_state_(&state_info, statep);	\
+	state = statep->ls_state = (s);				\
+} while (0)
 
-/*
+#define POP_STATE()	do {					\
+	if (--statep == state_info.base)			\
+		statep = pop_state_(&state_info, statep);	\
+	state = statep->ls_state;				\
+} while (0)
+
+/**
  * Lexical analyser
  *
  * tokens are not regular expressions, they are LL(1).
@@ -139,19 +187,17 @@ yylex(int cf)
 {
 	Lex_state states[STATE_BSIZE], *statep, *s2, *base;
 	State_info state_info;
-	int c, state;
+	int c, c2, state;
 	XString ws;		/* expandable output word */
 	char *wp;		/* output word pointer */
 	char *sp, *dp;
-	int c2;
-	bool last_terminal_was_bracket;
 
  Again:
 	states[0].ls_state = -1;
 	states[0].ls_info.base = NULL;
 	statep = &states[1];
 	state_info.base = states;
-	state_info.end = &states[STATE_BSIZE];
+	state_info.end = &state_info.base[STATE_BSIZE];
 
 	Xinit(ws, wp, 64, ATEMP);
 
@@ -161,12 +207,15 @@ yylex(int cf)
 	if (cf&ONEWORD)
 		state = SWORD;
 	else if (cf&LETEXPR) {
-		*wp++ = OQUOTE;	 /* enclose arguments in (double) quotes */
+		/* enclose arguments in (double) quotes */
+		*wp++ = OQUOTE;
 		state = SLETPAREN;
 		statep->ls_sletparen.nparen = 0;
+#ifndef MKSH_SMALL
 	} else if (cf&LETARRAY) {
 		state = SLETARRAY;
 		statep->ls_sletarray.nparen = 0;
+#endif
 	} else {		/* normal lexing */
 		state = (cf & HEREDELIM) ? SHEREDELIM : SBASE;
 		while ((c = getsc()) == ' ' || c == '\t')
@@ -226,7 +275,7 @@ yylex(int cf)
 			/* FALLTHROUGH */
 		case SBASE:
 			if (c == '[' && (cf & (VARASN|ARRAYVAR))) {
-				*wp = EOS; /* temporary */
+				*wp = EOS;	/* temporary */
 				if (is_wdvarname(Xstring(ws, wp), false)) {
 					char *p, *tmp;
 
@@ -244,9 +293,9 @@ yylex(int cf)
 						Source *s;
 
 						s = pushs(SREREAD,
-							  source->areap);
-						s->start = s->str
-							= s->u.freeme = tmp;
+						    source->areap);
+						s->start = s->str =
+						    s->u.freeme = tmp;
 						s->next = source;
 						source = s;
 					}
@@ -255,7 +304,7 @@ yylex(int cf)
 				*wp++ = c;
 				break;
 			}
-			/* FALLTHRU */
+			/* FALLTHROUGH */
  Sbase1:		/* includes *(...|...) pattern (*+?@!) */
 			if (c == '*' || c == '@' || c == '+' || c == '?' ||
 			    c == '!') {
@@ -268,20 +317,25 @@ yylex(int cf)
 				}
 				ungetsc(c2);
 			}
-			/* FALLTHRU */
+			/* FALLTHROUGH */
  Sbase2:		/* doesn't include *(...|...) pattern (*+?@!) */
 			switch (c) {
 			case '\\':
-				c = getsc();
-				if (c) /* trailing \ is lost */
-					*wp++ = QCHAR, *wp++ = c;
+ getsc_qchar:
+				if ((c = getsc())) {
+					/* trailing \ is lost */
+					*wp++ = QCHAR;
+					*wp++ = c;
+				}
 				break;
 			case '\'':
+ open_ssquote:
 				*wp++ = OQUOTE;
 				ignore_backslash_newline++;
 				PUSH_STATE(SSQUOTE);
 				break;
 			case '"':
+ open_sdquote:
 				*wp++ = OQUOTE;
 				PUSH_STATE(SDQUOTE);
 				break;
@@ -301,19 +355,25 @@ yylex(int cf)
 					/* FALLTHROUGH */
 				case '\\':
 				case '$': case '`':
-					*wp++ = QCHAR, *wp++ = c;
+ store_qchar:
+					*wp++ = QCHAR;
+					*wp++ = c;
 					break;
 				default:
  heredocquote:
 					Xcheck(ws, wp);
-					if (c) { /* trailing \ is lost */
-						*wp++ = CHAR, *wp++ = '\\';
-						*wp++ = CHAR, *wp++ = c;
+					if (c) {
+						/* trailing \ is lost */
+						*wp++ = CHAR;
+						*wp++ = '\\';
+						*wp++ = CHAR;
+						*wp++ = c;
 					}
 					break;
 				}
 				break;
 			case '$':
+ subst_dollar:
 				c = getsc();
 				if (c == '(') /*)*/ {
 					c = getsc();
@@ -337,7 +397,8 @@ yylex(int cf)
 					c = getsc();
 					/* allow :# and :% (ksh88 compat) */
 					if (c == ':') {
-						*wp++ = CHAR, *wp++ = c;
+						*wp++ = CHAR;
+						*wp++ = c;
 						c = getsc();
 						if (c == ':') {
 							*wp++ = CHAR;
@@ -369,7 +430,8 @@ yylex(int cf)
 							break;
 						}
 					} else if (c == '/') {
-						*wp++ = CHAR, *wp++ = c;
+						*wp++ = CHAR;
+						*wp++ = c;
 						if ((c = getsc()) == '/') {
 							*wp++ = ADELIM;
 							*wp++ = c;
@@ -391,7 +453,10 @@ yylex(int cf)
 						PUSH_STATE(STBRACE);
 					} else {
 						ungetsc(c);
-						PUSH_STATE(SBRACE);
+						if (state == SDQUOTE)
+							PUSH_STATE(SQBRACE);
+						else
+							PUSH_STATE(SBRACE);
 					}
 				} else if (ksh_isalphx(c)) {
 					*wp++ = OSUBST;
@@ -413,18 +478,27 @@ yylex(int cf)
 					*wp++ = '\0';
 					*wp++ = CSUBST;
 					*wp++ = 'X';
+				} else if (c == '\'' && (state == SBASE)) {
+					/* XXX which other states are valid? */
+					*wp++ = OQUOTE;
+					ignore_backslash_newline++;
+					PUSH_STATE(SEQUOTE);
+					statep->ls_sequote.got_NUL = false;
+					break;
 				} else {
-					*wp++ = CHAR, *wp++ = '$';
+					*wp++ = CHAR;
+					*wp++ = '$';
 					ungetsc(c);
 				}
 				break;
 			case '`':
+ subst_gravis:
 				PUSH_STATE(SBQUOTE);
 				*wp++ = COMSUB;
 				/* Need to know if we are inside double quotes
-				 * since sh/at&t-ksh translate the \" to " in
-				 * "`..\"..`".
-				 * This is not done in posix mode (section
+				 * since sh/AT&T-ksh translate the \" to " in
+				 * "`...\"...`".
+				 * This is not done in POSIX mode (section
 				 * 3.2.3, Double Quotes: "The backquote shall
 				 * retain its special meaning introducing the
 				 * other form of command substitution (see
@@ -464,9 +538,42 @@ yylex(int cf)
 					*wp++ = getsc();
 					break;
 				}
-				/* FALLTHRU */
+				/* FALLTHROUGH */
 			default:
-				*wp++ = CHAR, *wp++ = c;
+ store_char:
+				*wp++ = CHAR;
+				*wp++ = c;
+			}
+			break;
+
+		case SEQUOTE:
+			if (c == '\'') {
+				POP_STATE();
+				*wp++ = CQUOTE;
+				ignore_backslash_newline--;
+			} else if (c == '\\') {
+				if ((c2 = unbksl(true, s_get, s_put)) == -1)
+					c2 = s_get();
+				if (c2 == 0)
+					statep->ls_sequote.got_NUL = true;
+				if (!statep->ls_sequote.got_NUL) {
+					char ts[4];
+
+					if ((unsigned int)c2 < 0x100) {
+						*wp++ = QCHAR;
+						*wp++ = c2;
+					} else {
+						c = utf_wctomb(ts, c2 - 0x100);
+						ts[c] = 0;
+						for (c = 0; ts[c]; ++c) {
+							*wp++ = QCHAR;
+							*wp++ = ts[c];
+						}
+					}
+				}
+			} else if (!statep->ls_sequote.got_NUL) {
+				*wp++ = QCHAR;
+				*wp++ = c;
 			}
 			break;
 
@@ -475,8 +582,10 @@ yylex(int cf)
 				POP_STATE();
 				*wp++ = CQUOTE;
 				ignore_backslash_newline--;
-			} else
-				*wp++ = QCHAR, *wp++ = c;
+			} else {
+				*wp++ = QCHAR;
+				*wp++ = c;
+			}
 			break;
 
 		case SDQUOTE:
@@ -487,14 +596,14 @@ yylex(int cf)
 				goto Subst;
 			break;
 
-		case SCSPAREN: /* $( .. ) */
+		case SCSPAREN:	/* $( ... ) */
 			/* todo: deal with $(...) quoting properly
-			 * kludge to partly fake quoting inside $(..): doesn't
-			 * really work because nested $(..) or ${..} inside
+			 * kludge to partly fake quoting inside $(...): doesn't
+			 * really work because nested $(...) or ${...} inside
 			 * double quotes aren't dealt with.
 			 */
 			switch (statep->ls_scsparen.csstate) {
-			case 0: /* normal */
+			case 0:	/* normal */
 				switch (c) {
 				case '(':
 					statep->ls_scsparen.nparen++;
@@ -515,19 +624,19 @@ yylex(int cf)
 				}
 				break;
 
-			case 1: /* backslash in normal mode */
-			case 3: /* backslash in double quotes */
+			case 1:	/* backslash in normal mode */
+			case 3:	/* backslash in double quotes */
 				--statep->ls_scsparen.csstate;
 				break;
 
-			case 2: /* double quotes */
+			case 2:	/* double quotes */
 				if (c == '"')
 					statep->ls_scsparen.csstate = 0;
 				else if (c == '\\')
 					statep->ls_scsparen.csstate = 3;
 				break;
 
-			case 4: /* single quotes */
+			case 4:	/* single quotes */
 				if (c == '\'') {
 					statep->ls_scsparen.csstate = 0;
 					ignore_backslash_newline--;
@@ -536,15 +645,14 @@ yylex(int cf)
 			}
 			if (statep->ls_scsparen.nparen == 0) {
 				POP_STATE();
-				*wp++ = 0; /* end of COMSUB */
+				*wp++ = 0;	/* end of COMSUB */
 			} else
 				*wp++ = c;
 			break;
 
-		case SASPAREN: /* $(( .. )) */
-			/* todo: deal with $((...); (...)) properly */
+		case SASPAREN:	/* $(( ... )) */
 			/* XXX should nest using existing state machine
-			 * (embed "..", $(...), etc.) */
+			 * (embed "...", $(...), etc.) */
 			if (c == '(')
 				statep->ls_sasparen.nparen++;
 			else if (c == ')') {
@@ -553,7 +661,8 @@ yylex(int cf)
 					/*(*/
 					if ((c2 = getsc()) == ')') {
 						POP_STATE();
-						*wp++ = 0; /* end of EXPRSUB */
+						/* end of EXPRSUB */
+						*wp++ = 0;
 						break;
 					} else {
 						char *s;
@@ -561,7 +670,7 @@ yylex(int cf)
 						ungetsc(c2);
 						/* mismatched parenthesis -
 						 * assume we were really
-						 * parsing a $(..) expression
+						 * parsing a $(...) expression
 						 */
 						s = Xrestpos(ws, wp,
 						    statep->ls_sasparen.start);
@@ -579,20 +688,45 @@ yylex(int cf)
 			*wp++ = c;
 			break;
 
+		case SQBRACE:
+			if (c == '\\') {
+				/*
+				 * perform POSIX "quote removal" if the back-
+				 * slash is "special", i.e. same cases as the
+				 * {case '\\':} in Subst: plus closing brace;
+				 * in mksh code "quote removal" on '\c' means
+				 * write QCHAR+c, otherwise CHAR+\+CHAR+c are
+				 * emitted (in heredocquote:)
+				 */
+				if ((c = getsc()) == '"' || c == '\\' ||
+				    c == '$' || c == '`' || c == /*{*/'}')
+					goto store_qchar;
+				goto heredocquote;
+			}
+			goto common_SQBRACE;
+
 		case SBRACE:
-			/*{*/
-			if (c == '}') {
-				POP_STATE();
-				*wp++ = CSUBST;
-				*wp++ = /*{*/ '}';
-			} else
-				goto Sbase1;
+			if (c == '\'')
+				goto open_ssquote;
+			else if (c == '\\')
+				goto getsc_qchar;
+ common_SQBRACE:
+			if (c == '"')
+				goto open_sdquote;
+			else if (c == '$')
+				goto subst_dollar;
+			else if (c == '`')
+				goto subst_gravis;
+			else if (c != /*{*/ '}')
+				goto store_char;
+			POP_STATE();
+			*wp++ = CSUBST;
+			*wp++ = /*{*/ '}';
 			break;
 
 		case STBRACE:
-			/* Same as SBRACE, except (,|,) treated specially */
-			/*{*/
-			if (c == '}') {
+			/* Same as SBASE, except (,|,) treated specially */
+			if (c == /*{*/ '}') {
 				POP_STATE();
 				*wp++ = CSUBST;
 				*wp++ = /*{*/ '}';
@@ -621,9 +755,10 @@ yylex(int cf)
 						*wp++ = c;
 						break;
 					}
-					/* FALLTHRU */
+					/* FALLTHROUGH */
 				default:
-					if (c) { /* trailing \ is lost */
+					if (c) {
+						/* trailing \ is lost */
 						*wp++ = '\\';
 						*wp++ = c;
 					}
@@ -640,22 +775,37 @@ yylex(int cf)
 			/*(*/
 			if (c == ')') {
 				if (statep->ls_sletparen.nparen > 0)
-				    --statep->ls_sletparen.nparen;
-				/*(*/
-				else if ((c2 = getsc()) == ')') {
+					--statep->ls_sletparen.nparen;
+				else if ((c2 = getsc()) == /*(*/ ')') {
 					c = 0;
 					*wp++ = CQUOTE;
 					goto Done;
-				} else
+				} else {
+					Source *s;
+
 					ungetsc(c2);
+					/* mismatched parenthesis -
+					 * assume we were really
+					 * parsing a $(...) expression
+					 */
+					*wp = EOS;
+					sp = Xstring(ws, wp);
+					dp = wdstrip(sp, true, false);
+					s = pushs(SREREAD, source->areap);
+					s->start = s->str = s->u.freeme = dp;
+					s->next = source;
+					source = s;
+					return ('('/*)*/);
+				}
 			} else if (c == '(')
 				/* parenthesis inside quotes and backslashes
-				 * are lost, but at&t ksh doesn't count them
+				 * are lost, but AT&T ksh doesn't count them
 				 * either
 				 */
 				++statep->ls_sletparen.nparen;
 			goto Sbase2;
 
+#ifndef MKSH_SMALL
 		case SLETARRAY:	/* LETARRAY: =( ... ) */
 			if (c == '('/*)*/)
 				++statep->ls_sletarray.nparen;
@@ -664,20 +814,32 @@ yylex(int cf)
 					c = 0;
 					goto Done;
 				}
-			*wp++ = CHAR, *wp++ = c;
+			*wp++ = CHAR;
+			*wp++ = c;
 			break;
+#endif
 
 		case SHERESTRING:	/* <<< delimiter */
 			if (c == '\\') {
 				c = getsc();
-				if (c) { /* trailing \ is lost */
+				if (c) {
+					/* trailing \ is lost */
 					*wp++ = QCHAR;
 					*wp++ = c;
 				}
 				/* invoke quoting mode */
 				Xstring(ws, wp)[0] = QCHAR;
+			} else if (c == '$') {
+				if ((c2 = getsc()) == '\'') {
+					PUSH_STATE(SEQUOTE);
+					statep->ls_sequote.got_NUL = false;
+					goto sherestring_quoted;
+				}
+				ungetsc(c2);
+				goto sherestring_regular;
 			} else if (c == '\'') {
 				PUSH_STATE(SSQUOTE);
+ sherestring_quoted:
 				*wp++ = OQUOTE;
 				ignore_backslash_newline++;
 				/* invoke quoting mode */
@@ -687,6 +849,7 @@ yylex(int cf)
 				*wp++ = OQUOTE;
 				/* just don't IFS split; no quoting mode */
 			} else {
+ sherestring_regular:
 				*wp++ = CHAR;
 				*wp++ = c;
 			}
@@ -694,27 +857,38 @@ yylex(int cf)
 
 		case SHEREDELIM:	/* <<,<<- delimiter */
 			/* XXX chuck this state (and the next) - use
-			 * the existing states ($ and \`..` should be
+			 * the existing states ($ and \`...` should be
 			 * stripped of their specialness after the
 			 * fact).
 			 */
 			/* here delimiters need a special case since
-			 * $ and `..` are not to be treated specially
+			 * $ and `...` are not to be treated specially
 			 */
 			if (c == '\\') {
 				c = getsc();
-				if (c) { /* trailing \ is lost */
+				if (c) {
+					/* trailing \ is lost */
 					*wp++ = QCHAR;
 					*wp++ = c;
 				}
+			} else if (c == '$') {
+				if ((c2 = getsc()) == '\'') {
+					PUSH_STATE(SEQUOTE);
+					statep->ls_sequote.got_NUL = false;
+					goto sheredelim_quoted;
+				}
+				ungetsc(c2);
+				goto sheredelim_regular;
 			} else if (c == '\'') {
 				PUSH_STATE(SSQUOTE);
+ sheredelim_quoted:
 				*wp++ = OQUOTE;
 				ignore_backslash_newline++;
 			} else if (c == '"') {
 				state = statep->ls_state = SHEREDQUOTE;
 				*wp++ = OQUOTE;
 			} else {
+ sheredelim_regular:
 				*wp++ = CHAR;
 				*wp++ = c;
 			}
@@ -734,7 +908,8 @@ yylex(int cf)
 					case '$': case '`':
 						break;
 					default:
-						if (c) { /* trailing \ lost */
+						if (c) {
+							/* trailing \ lost */
 							*wp++ = CHAR;
 							*wp++ = '\\';
 						}
@@ -767,8 +942,10 @@ yylex(int cf)
 		/* XXX figure out what is missing */
 		yyerror("no closing quote\n");
 
+#ifndef MKSH_SMALL
 	if (state == SLETARRAY && statep->ls_sletarray.nparen != -1)
 		yyerror("%s: ')' missing\n", T_synerr);
+#endif
 
 	/* This done to avoid tests for SHEREDELIM wherever SBASE tested */
 	if (state == SHEREDELIM || state == SHERESTRING)
@@ -776,8 +953,7 @@ yylex(int cf)
 
 	dp = Xstring(ws, wp);
 	if ((c == '<' || c == '>' || c == '&') && state == SBASE) {
-		struct ioword *iop = (struct ioword *)alloc(sizeof (*iop),
-		    ATEMP);
+		struct ioword *iop = alloc(sizeof(struct ioword), ATEMP);
 
 		if (Xlength(ws, wp) == 0)
 			iop->unit = c == '<' ? 0 : 1;
@@ -828,7 +1004,7 @@ yylex(int cf)
 		iop->heredoc = NULL;
 		Xfree(ws, wp);	/* free word */
 		yylval.iop = iop;
-		return REDIR;
+		return (REDIR);
  no_iop:
 		;
 	}
@@ -841,11 +1017,6 @@ yylex(int cf)
 				c = (c == ';') ? BREAK :
 				    (c == '|') ? LOGOR :
 				    (c == '&') ? LOGAND :
-				    /*
-				     * this is the place where
-				     * ((...); (...))
-				     * and similar is broken
-				     */
 				    /* c == '(' ) */ MDPAREN;
 			else if (c == '|' && c2 == '&')
 				c = COPROC;
@@ -863,12 +1034,22 @@ yylex(int cf)
 
 	*wp++ = EOS;		/* terminate word */
 	yylval.cp = Xclose(ws, wp);
-	if (state == SWORD || state == SLETPAREN ||
-	    state == SLETARRAY)	/* ONEWORD? */
-		return LWORD;
+	if (state == SWORD || state == SLETPAREN
+	    /* XXX ONEWORD? */
+#ifndef MKSH_SMALL
+	    || state == SLETARRAY
+#endif
+	    )
+		return (LWORD);
 
-	last_terminal_was_bracket = c == '(';
-	ungetsc(c);		/* unget terminator */
+	/* unget terminator */
+	ungetsc(c);
+
+	/*
+	 * note: the alias-vs-function code below depends on several
+	 * interna: starting from here, source->str is not modified;
+	 * the way getsc() and ungetsc() operate; etc.
+	 */
 
 	/* copy word to unprefixed string ident */
 	sp = yylval.cp;
@@ -890,30 +1071,61 @@ yylex(int cf)
 
 	if (*ident != '\0' && (cf&(KEYWORD|ALIAS))) {
 		struct tbl *p;
-		int h = hash(ident);
+		uint32_t h = hash(ident);
 
 		/* { */
 		if ((cf & KEYWORD) && (p = ktsearch(&keywords, ident, h)) &&
 		    (!(cf & ESACONLY) || p->val.i == ESAC || p->val.i == '}')) {
 			afree(yylval.cp, ATEMP);
-			return p->val.i;
+			return (p->val.i);
 		}
 		if ((cf & ALIAS) && (p = ktsearch(&aliases, ident, h)) &&
 		    (p->flag & ISSET)) {
-			if (last_terminal_was_bracket)
-				/* prefer functions over aliases */
+			/*
+			 * this still points to the same character as the
+			 * ungetsc'd terminator from above
+			 */
+			const char *cp = source->str;
+
+			/* prefer POSIX but not Korn functions over aliases */
+			while (*cp == ' ' || *cp == '\t')
+				/*
+				 * this is like getsc() without skipping
+				 * over Source boundaries (including not
+				 * parsing ungetsc'd characters that got
+				 * pushed into an SREREAD) which is what
+				 * we want here anyway: find out whether
+				 * the alias name is followed by a POSIX
+				 * function definition (only the opening
+				 * parenthesis is checked though)
+				 */
+				++cp;
+			/* prefer functions over aliases */
+			if (*cp == '(' /*)*/)
+				/*
+				 * delete alias upon encountering function
+				 * definition
+				 */
 				ktdelete(p);
 			else {
-				Source *s;
+				Source *s = source;
 
-				for (s = source; s->type == SALIAS; s = s->next)
+				while (s && (s->flags & SF_HASALIAS))
 					if (s->u.tblp == p)
-						return LWORD;
+						return (LWORD);
+					else
+						s = s->next;
 				/* push alias expansion */
 				s = pushs(SALIAS, source->areap);
 				s->start = s->str = p->val.s;
 				s->u.tblp = p;
+				s->flags |= SF_HASALIAS;
 				s->next = source;
+				if (source->type == SEOF) {
+					/* prevent infinite recursion at EOS */
+					source->u.tblp = p;
+					source->flags |= SF_HASALIAS;
+				}
 				source = s;
 				afree(yylval.cp, ATEMP);
 				goto Again;
@@ -921,7 +1133,7 @@ yylex(int cf)
 		}
 	}
 
-	return LWORD;
+	return (LWORD);
 }
 
 static void
@@ -1035,21 +1247,14 @@ pushs(int type, Area *areap)
 {
 	Source *s;
 
-	s = (Source *)alloc(sizeof (Source), areap);
+	s = alloc(sizeof(Source), areap);
+	memset(s, 0, sizeof(Source));
 	s->type = type;
 	s->str = null;
-	s->start = NULL;
-	s->line = 0;
-	s->errline = 0;
-	s->file = NULL;
-	s->flags = 0;
-	s->next = NULL;
 	s->areap = areap;
 	if (type == SFILE || type == SSTDIN)
 		XinitN(s->xs, 256, s->areap);
-	else
-		memset(&s->xs, 0, sizeof(s->xs));
-	return s;
+	return (s);
 }
 
 static int
@@ -1064,7 +1269,7 @@ getsc__(void)
 		switch (s->type) {
 		case SEOF:
 			s->str = null;
-			return 0;
+			return (0);
 
 		case SSTDIN:
 		case SFILE:
@@ -1099,7 +1304,7 @@ getsc__(void)
 				source->flags |= s->flags & SF_ALIAS;
 				s = source;
 			} else if (*s->u.tblp->val.s &&
-			    ksh_isspace(strnul(s->u.tblp->val.s)[-1])) {
+			    (c = strnul(s->u.tblp->val.s)[-1], ksh_isspace(c))) {
 				source = s = s->next;	/* pop source stack */
 				/* Note that this alias ended with a space,
 				 * enabling alias expansion on the following
@@ -1110,7 +1315,7 @@ getsc__(void)
 				/* At this point, we need to keep the current
 				 * alias in the source list so recursive
 				 * aliases can be detected and we also need
-				 * to return the next character.  Do this
+				 * to return the next character. Do this
 				 * by temporarily popping the alias to get
 				 * the next character and then put it back
 				 * in the source list with the SF_ALIASEND
@@ -1135,7 +1340,7 @@ getsc__(void)
 			continue;
 
 		case SREREAD:
-			if (s->start != s->ugbuf) /* yuck */
+			if (s->start != s->ugbuf)	/* yuck */
 				afree(s->u.freeme, ATEMP);
 			source = s = s->next;
 			continue;
@@ -1143,7 +1348,7 @@ getsc__(void)
 		if (s->str == NULL) {
 			s->type = SEOF;
 			s->start = s->str = null;
-			return '\0';
+			return ('\0');
 		}
 		if (s->flags & SF_ECHO) {
 			shf_puts(s->str, shl_out);
@@ -1157,18 +1362,18 @@ getsc__(void)
 		    (((const unsigned char *)(s->str))[0] == 0xBB) &&
 		    (((const unsigned char *)(s->str))[1] == 0xBF)) {
 			s->str += 2;
-			Flag(FUTFHACK) = 1;
+			UTFMODE = 1;
 			goto getsc_again;
 		}
 	}
-	return c;
+	return (c);
 }
 
 static void
 getsc_line(Source *s)
 {
-	char *xp = Xstring(s->xs, xp);
-	int interactive = Flag(FTALKING) && s->type == SSTDIN;
+	char *xp = Xstring(s->xs, xp), *cp;
+	bool interactive = Flag(FTALKING) && s->type == SSTDIN;
 	int have_tty = interactive && (s->flags & SF_TTY);
 
 	/* Done here to ensure nothing odd happens when a timeout occurs */
@@ -1181,7 +1386,7 @@ getsc_line(Source *s)
 		alarm(ksh_tmout);
 	}
 	if (have_tty && (
-#ifndef MKSH_NOVI
+#if !MKSH_S_NOVI
 	    Flag(FVI) ||
 #endif
 	    Flag(FEMACS) || Flag(FGMACS))) {
@@ -1194,6 +1399,8 @@ getsc_line(Source *s)
 		xp += nread;
 	} else {
 		if (interactive) {
+			if (got_winch)
+				change_winsz();
 			pprompt(prompt, 0);
 		} else
 			s->line++;
@@ -1211,14 +1418,14 @@ getsc_line(Source *s)
 			if (!p || (xp = p, xp[-1] == '\n'))
 				break;
 			/* double buffer size */
-			xp++; /* move past null so doubling works... */
+			xp++;	/* move past NUL so doubling works... */
 			XcheckN(s->xs, xp, Xlength(s->xs, xp));
-			xp--; /* ...and move back again */
+			xp--;	/* ...and move back again */
 		}
 		/* flush any unwanted input so other programs/builtins
-		 * can read it.  Not very optimal, but less error prone
+		 * can read it. Not very optimal, but less error prone
 		 * than flushing else where, dealing with redirections,
-		 * etc..
+		 * etc.
 		 * todo: reduce size of shf buffer (~128?) if SSTDIN
 		 */
 		if (s->type == SSTDIN)
@@ -1232,22 +1439,45 @@ getsc_line(Source *s)
 		ksh_tmout_state = TMOUT_EXECUTING;
 		alarm(0);
 	}
-	s->start = s->str = Xstring(s->xs, xp);
+	cp = Xstring(s->xs, xp);
+#ifndef MKSH_SMALL
+	if (interactive && *cp == '!' && cur_prompt == PS1) {
+		int linelen;
+
+		linelen = Xlength(s->xs, xp);
+		XcheckN(s->xs, xp, fc_e_n + /* NUL */ 1);
+		/* reload after potential realloc */
+		cp = Xstring(s->xs, xp);
+		/* change initial '!' into space */
+		*cp = ' ';
+		/* NUL terminate the current string */
+		*xp = '\0';
+		/* move the actual string forward */
+		memmove(cp + fc_e_n, cp, linelen + /* NUL */ 1);
+		xp += fc_e_n;
+		/* prepend it with "fc -e -" */
+		memcpy(cp, fc_e_, fc_e_n);
+	}
+#endif
+	s->start = s->str = cp;
 	strip_nuls(Xstring(s->xs, xp), Xlength(s->xs, xp));
 	/* Note: if input is all nulls, this is not eof */
-	if (Xlength(s->xs, xp) == 0) { /* EOF */
+	if (Xlength(s->xs, xp) == 0) {
+		/* EOF */
 		if (s->type == SFILE)
 			shf_fdclose(s->u.shf);
 		s->str = NULL;
-	} else if (interactive) {
-		char *p = Xstring(s->xs, xp);
-		if (cur_prompt == PS1)
-			while (*p && ctype(*p, C_IFS) && ctype(*p, C_IFSWS))
-				p++;
-		if (*p) {
-			s->line++;
-			histsave(s->line, s->str, 1);
-		}
+	} else if (interactive && *s->str &&
+	    (cur_prompt != PS1 || !ctype(*s->str, C_IFS | C_IFSWS))) {
+		histsave(&s->line, s->str, true, true);
+#if !defined(MKSH_SMALL) && HAVE_PERSISTENT_HISTORY
+	} else if (interactive && cur_prompt == PS1) {
+		cp = Xstring(s->xs, xp);
+		while (*cp && ctype(*cp, C_IFSWS))
+			++cp;
+		if (!*cp)
+			histsync();
+#endif
 	}
 	if (interactive)
 		set_prompt(PS2, NULL);
@@ -1259,10 +1489,10 @@ set_prompt(int to, Source *s)
 	cur_prompt = to;
 
 	switch (to) {
-	case PS1: /* command */
+	case PS1:	/* command */
 		/* Substitute ! and !! here, before substitutions are done
 		 * so ! in expanded variables are not expanded.
-		 * NOTE: this is not what at&t ksh does (it does it after
+		 * NOTE: this is not what AT&T ksh does (it does it after
 		 * substitutions, POSIX doesn't say which is to be done.
 		 */
 		{
@@ -1285,26 +1515,26 @@ set_prompt(int to, Source *s)
 			if (sigsetjmp(e->jbuf, 0)) {
 				prompt = safe_prompt;
 				/* Don't print an error - assume it has already
-				 * been printed.  Reason is we may have forked
+				 * been printed. Reason is we may have forked
 				 * to run a command and the child may be
 				 * unwinding its stack through this code as it
 				 * exits.
 				 */
 			} else {
 				char *cp = substitute(ps1, 0);
-				prompt = str_save(cp, saved_atemp);
+				strdupx(prompt, cp, saved_atemp);
 			}
 			quitenv(NULL);
 		}
 		break;
-	case PS2: /* command continuation */
+	case PS2:	/* command continuation */
 		prompt = str_val(global("PS2"));
 		break;
 	}
 }
 
 static int
-dopprompt(const char *cp, int ntruncate, int doprint)
+dopprompt(const char *cp, int ntruncate, bool doprint)
 {
 	int columns = 0, lines = 0, indelimit = 0;
 	char delimiter = 0;
@@ -1332,7 +1562,7 @@ dopprompt(const char *cp, int ntruncate, int doprint)
 				columns--;
 		} else if (*cp == delimiter)
 			indelimit = !indelimit;
-		else if (Flag(FUTFHACK) && ((unsigned char)*cp > 0x7F)) {
+		else if (UTFMODE && ((unsigned char)*cp > 0x7F)) {
 			const char *cp2;
 			columns += utf_widthadj(cp, &cp2);
 			if (doprint && (indelimit ||
@@ -1348,21 +1578,20 @@ dopprompt(const char *cp, int ntruncate, int doprint)
 	}
 	if (doprint)
 		shf_flush(shl_out);
-	indelimit = (x_cols * lines + columns);
-	return indelimit;
+	return (x_cols * lines + columns);
 }
 
 
 void
 pprompt(const char *cp, int ntruncate)
 {
-	dopprompt(cp, ntruncate, 1);
+	dopprompt(cp, ntruncate, true);
 }
 
 int
 promptlen(const char *cp)
 {
-	return (dopprompt(cp, 0, 0));
+	return (dopprompt(cp, 0, false));
 }
 
 /* Read the variable part of a ${...} expression (ie, up to but not including
@@ -1372,10 +1601,9 @@ static char *
 get_brace_var(XString *wsp, char *wp)
 {
 	enum parse_state {
-			   PS_INITIAL, PS_SAW_HASH, PS_IDENT,
-			   PS_NUMBER, PS_VAR1, PS_END
-			 }
-		state;
+		PS_INITIAL, PS_SAW_HASH, PS_IDENT,
+		PS_NUMBER, PS_VAR1
+	} state;
 	char c;
 
 	state = PS_INITIAL;
@@ -1384,11 +1612,11 @@ get_brace_var(XString *wsp, char *wp)
 		/* State machine to figure out where the variable part ends. */
 		switch (state) {
 		case PS_INITIAL:
-			if (c == '#') {
+			if (c == '#' || c == '!' || c == '%') {
 				state = PS_SAW_HASH;
 				break;
 			}
-			/* FALLTHRU */
+			/* FALLTHROUGH */
 		case PS_SAW_HASH:
 			if (ksh_isalphx(c))
 				state = PS_IDENT;
@@ -1397,11 +1625,10 @@ get_brace_var(XString *wsp, char *wp)
 			else if (ctype(c, C_VAR1))
 				state = PS_VAR1;
 			else
-				state = PS_END;
+				goto out;
 			break;
 		case PS_IDENT:
 			if (!ksh_isalnux(c)) {
-				state = PS_END;
 				if (c == '[') {
 					char *tmp, *p;
 
@@ -1413,29 +1640,25 @@ get_brace_var(XString *wsp, char *wp)
 						*wp++ = *p++;
 					}
 					afree(tmp, ATEMP);
-					c = getsc(); /* the ] */
+					c = getsc();	/* the ] */
 				}
+				goto out;
 			}
 			break;
 		case PS_NUMBER:
 			if (!ksh_isdigit(c))
-				state = PS_END;
+				goto out;
 			break;
 		case PS_VAR1:
-			state = PS_END;
-			break;
-		case PS_END: /* keep gcc happy */
-			break;
-		}
-		if (state == PS_END) {
-			*wp++ = '\0';	/* end of variable part */
-			ungetsc(c);
-			break;
+			goto out;
 		}
 		Xcheck(*wsp, wp);
 		*wp++ = c;
 	}
-	return wp;
+ out:
+	*wp++ = '\0';	/* end of variable part */
+	ungetsc(c);
+	return (wp);
 }
 
 /*
@@ -1466,7 +1689,7 @@ arraysub(char **strp)
 	*wp++ = '\0';
 	*strp = Xclose(ws, wp);
 
-	return depth == 0 ? 1 : 0;
+	return (depth == 0 ? 1 : 0);
 }
 
 /* Unget a char: handles case when we are already at the start of the buffer */
@@ -1477,7 +1700,7 @@ ungetsc(int c)
 		backslash_skip--;
 	/* Don't unget eof... */
 	if (source->str == null && c == '\0')
-		return source->str;
+		return (source->str);
 	if (source->str > source->start)
 		source->str--;
 	else {
@@ -1489,7 +1712,7 @@ ungetsc(int c)
 		s->next = source;
 		source = s;
 	}
-	return source->str;
+	return (source->str);
 }
 
 
@@ -1500,11 +1723,11 @@ getsc_bn(void)
 	int c, c2;
 
 	if (ignore_backslash_newline)
-		return getsc_();
+		return (getsc_());
 
 	if (backslash_skip == 1) {
 		backslash_skip = 2;
-		return getsc_();
+		return (getsc_());
 	}
 
 	backslash_skip = 0;
@@ -1518,19 +1741,19 @@ getsc_bn(void)
 			ungetsc(c2);
 			backslash_skip = 1;
 		}
-		return c;
+		return (c);
 	}
 }
 
 static Lex_state *
 push_state_(State_info *si, Lex_state *old_end)
 {
-	Lex_state *new = alloc(sizeof (Lex_state) * STATE_BSIZE, ATEMP);
+	Lex_state *news = alloc(STATE_BSIZE * sizeof(Lex_state), ATEMP);
 
-	new[0].ls_info.base = old_end;
-	si->base = &new[0];
-	si->end = &new[STATE_BSIZE];
-	return &new[1];
+	news[0].ls_info.base = old_end;
+	si->base = &news[0];
+	si->end = &news[STATE_BSIZE];
+	return (&news[1]);
 }
 
 static Lex_state *
@@ -1543,5 +1766,17 @@ pop_state_(State_info *si, Lex_state *old_end)
 
 	afree(old_base, ATEMP);
 
-	return si->base + STATE_BSIZE - 1;
+	return (si->base + STATE_BSIZE - 1);
+}
+
+static int
+s_get(void)
+{
+	return (getsc());
+}
+
+static void
+s_put(int c)
+{
+	ungetsc(c);
 }
