@@ -1,12 +1,32 @@
-/*	$OpenBSD: tree.c,v 1.18 2006/04/10 14:38:59 jaredy Exp $	*/
+/*	$OpenBSD: tree.c,v 1.19 2008/08/11 21:50:35 jaredy Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/tree.c,v 1.18 2008/07/09 21:32:45 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/tree.c,v 1.30 2010/02/25 20:18:19 tg Exp $");
 
 #define INDENT	4
 
-#define tputc(c, shf)	shf_putchar(c, shf);
+#define tputc(c, shf) shf_putchar(c, shf);
 static void ptree(struct op *, int, struct shf *);
 static void pioact(struct shf *, int, struct ioword *);
 static void tputC(int, struct shf *);
@@ -64,8 +84,7 @@ ptree(struct op *t, int indent, struct shf *shf)
 		shf_puts("! ", shf);
 		t = t->right;
 		goto Chain;
-	case TDBRACKET:
-	  {
+	case TDBRACKET: {
 		int i;
 
 		shf_puts("[[", shf);
@@ -73,10 +92,10 @@ ptree(struct op *t, int indent, struct shf *shf)
 			fptreef(shf, indent, " %S", t->args[i]);
 		shf_puts(" ]] ", shf);
 		break;
-	  }
+	}
 	case TSELECT:
 		fptreef(shf, indent, "select %s ", t->str);
-		/* FALLTHRU */
+		/* FALLTHROUGH */
 	case TFOR:
 		if (t->type == TFOR)
 			fptreef(shf, indent, "for %s ", t->str);
@@ -252,7 +271,7 @@ tputC(int c, struct shf *shf)
 static void
 tputS(char *wp, struct shf *shf)
 {
-	int c, quoted=0;
+	int c, quotelevel = 0;
 
 	/* problems:
 	 *	`...` -> $(...)
@@ -262,7 +281,7 @@ tputS(char *wp, struct shf *shf)
 	 *	COMSUB [(`] ...\0	(handle $ ` \ and maybe " in `...` case)
 	 */
 	while (1)
-		switch ((c = *wp++)) {
+		switch (*wp++) {
 		case EOS:
 			return;
 		case ADELIM:
@@ -271,7 +290,7 @@ tputS(char *wp, struct shf *shf)
 			break;
 		case QCHAR:
 			c = *wp++;
-			if (!quoted || (c == '"' || c == '`' || c == '$'))
+			if (!quotelevel || (c == '"' || c == '`' || c == '$'))
 				tputc('\\', shf);
 			tputC(c, shf);
 			break;
@@ -290,11 +309,12 @@ tputS(char *wp, struct shf *shf)
 			wp++;
 			break;
 		case OQUOTE:
-			quoted = 1;
+			quotelevel++;
 			tputc('"', shf);
 			break;
 		case CQUOTE:
-			quoted = 0;
+			if (quotelevel)
+				quotelevel--;
 			tputc('"', shf);
 			break;
 		case OSUBST:
@@ -419,22 +439,25 @@ tcopy(struct op *t, Area *ap)
 	char **rw;
 
 	if (t == NULL)
-		return NULL;
+		return (NULL);
 
-	r = (struct op *)alloc(sizeof (struct op), ap);
+	r = alloc(sizeof(struct op), ap);
 
 	r->type = t->type;
 	r->u.evalflags = t->u.evalflags;
 
-	r->str = t->type == TCASE ? wdcopy(t->str, ap) : str_save(t->str, ap);
+	if (t->type == TCASE)
+		r->str = wdcopy(t->str, ap);
+	else
+		strdupx(r->str, t->str, ap);
 
 	if (t->vars == NULL)
 		r->vars = NULL;
 	else {
 		for (tw = (const char **)t->vars; *tw++ != NULL; )
 			;
-		rw = r->vars = (char **)alloc((tw -
-		    (const char **)t->vars + 1) * sizeof (*tw), ap);
+		rw = r->vars = alloc((tw - (const char **)t->vars + 1) *
+		    sizeof(*tw), ap);
 		for (tw = (const char **)t->vars; *tw != NULL; )
 			*rw++ = wdcopy(*tw++, ap);
 		*rw = NULL;
@@ -445,8 +468,8 @@ tcopy(struct op *t, Area *ap)
 	else {
 		for (tw = t->args; *tw++ != NULL; )
 			;
-		r->args = (const char **)(rw = (char **)alloc((tw -
-		    t->args + 1) * sizeof (*tw), ap));
+		r->args = (const char **)(rw = alloc((tw - t->args + 1) *
+		    sizeof(*tw), ap));
 		for (tw = t->args; *tw != NULL; )
 			*rw++ = wdcopy(*tw++, ap);
 		*rw = NULL;
@@ -458,14 +481,14 @@ tcopy(struct op *t, Area *ap)
 	r->right = tcopy(t->right, ap);
 	r->lineno = t->lineno;
 
-	return r;
+	return (r);
 }
 
 char *
 wdcopy(const char *wp, Area *ap)
 {
 	size_t len = wdscan(wp, EOS) - wp;
-	return memcpy(alloc(len, ap), wp, len);
+	return (memcpy(alloc(len, ap), wp, len));
 }
 
 /* return the position of prefix c in wp plus 1 */
@@ -541,9 +564,9 @@ wdstrip(const char *wp, bool keepq, bool make_magic)
 	 *	x${foo:-'hi'} -> x${foo:-hi} unless keepq
 	 */
 	while (1)
-		switch ((c = *wp++)) {
+		switch (*wp++) {
 		case EOS:
-			return shf_sclose(&shf); /* null terminates */
+			return (shf_sclose(&shf)); /* null terminates */
 		case ADELIM:
 		case CHAR:
 			c = *wp++;
@@ -615,13 +638,13 @@ iocopy(struct ioword **iow, Area *ap)
 
 	for (ior = iow; *ior++ != NULL; )
 		;
-	ior = (struct ioword **)alloc((ior - iow + 1) * sizeof (*ior), ap);
+	ior = alloc((ior - iow + 1) * sizeof(struct ioword *), ap);
 
 	for (i = 0; iow[i] != NULL; i++) {
 		struct ioword *p, *q;
 
 		p = iow[i];
-		q = (struct ioword *)alloc(sizeof (*p), ap);
+		q = alloc(sizeof(struct ioword), ap);
 		ior[i] = q;
 		*q = *p;
 		if (p->name != NULL)
@@ -629,11 +652,11 @@ iocopy(struct ioword **iow, Area *ap)
 		if (p->delim != NULL)
 			q->delim = wdcopy(p->delim, ap);
 		if (p->heredoc != NULL)
-			q->heredoc = str_save(p->heredoc, ap);
+			strdupx(q->heredoc, p->heredoc, ap);
 	}
 	ior[i] = NULL;
 
-	return ior;
+	return (ior);
 }
 
 /*
@@ -648,12 +671,12 @@ tfree(struct op *t, Area *ap)
 		return;
 
 	if (t->str != NULL)
-		afree((void*)t->str, ap);
+		afree(t->str, ap);
 
 	if (t->vars != NULL) {
 		for (w = t->vars; *w != NULL; w++)
-			afree((void*)*w, ap);
-		afree((void*)t->vars, ap);
+			afree(*w, ap);
+		afree(t->vars, ap);
 	}
 
 	if (t->args != NULL) {
@@ -661,8 +684,8 @@ tfree(struct op *t, Area *ap)
 		/* XXX we assume the caller is right */
 		cw.ro = t->args;
 		for (w = cw.rw; *w != NULL; w++)
-			afree((void*)*w, ap);
-		afree((void*)t->args, ap);
+			afree(*w, ap);
+		afree(t->args, ap);
 	}
 
 	if (t->ioact != NULL)
@@ -671,7 +694,7 @@ tfree(struct op *t, Area *ap)
 	tfree(t->left, ap);
 	tfree(t->right, ap);
 
-	afree((void*)t, ap);
+	afree(t, ap);
 }
 
 static void
@@ -682,11 +705,12 @@ iofree(struct ioword **iow, Area *ap)
 
 	for (iop = iow; (p = *iop++) != NULL; ) {
 		if (p->name != NULL)
-			afree((void*)p->name, ap);
+			afree(p->name, ap);
 		if (p->delim != NULL)
-			afree((void*)p->delim, ap);
+			afree(p->delim, ap);
 		if (p->heredoc != NULL)
-			afree((void*)p->heredoc, ap);
-		afree((void*)p, ap);
+			afree(p->heredoc, ap);
+		afree(p, ap);
 	}
+	afree(iow, ap);
 }
