@@ -1,8 +1,32 @@
 /*	$OpenBSD: var.c,v 1.34 2007/10/15 02:16:35 deraadt Exp $	*/
 
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
+
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/var.c,v 1.57 2008/07/12 16:56:40 tg Exp $");
+#if defined(__OpenBSD__)
+#include <sys/sysctl.h>
+#endif
+
+__RCSID("$MirOS: src/bin/mksh/var.c,v 1.109 2010/07/18 17:29:50 tg Exp $");
 
 /*
  * Variables
@@ -15,17 +39,22 @@ __RCSID("$MirOS: src/bin/mksh/var.c,v 1.57 2008/07/12 16:56:40 tg Exp $");
  */
 static struct tbl vtemp;
 static struct table specials;
-static char	*formatstr(struct tbl *, const char *);
-static void	export(struct tbl *, const char *);
-static int	special(const char *);
-static void	unspecial(const char *);
-static void	getspec(struct tbl *);
-static void	setspec(struct tbl *);
-static void	unsetspec(struct tbl *);
+static char *formatstr(struct tbl *, const char *);
+static void exportprep(struct tbl *, const char *);
+static int special(const char *);
+static void unspecial(const char *);
+static void getspec(struct tbl *);
+static void setspec(struct tbl *);
+static void unsetspec(struct tbl *);
+static int getint(struct tbl *, mksh_ari_t *, bool);
+static mksh_ari_t intval(struct tbl *);
 static struct tbl *arraysearch(struct tbl *, uint32_t);
 static const char *array_index_calc(const char *, bool *, uint32_t *);
-static int rnd_get(void);
-static void rnd_set(unsigned long);
+static uint32_t oaathash_update(register uint32_t, register const uint8_t *,
+    register size_t);
+static uint32_t oaathash_finalise(register uint32_t);
+
+uint8_t set_refflag = 0;
 
 /*
  * create a new block for function calls and simple commands
@@ -37,7 +66,7 @@ newblock(void)
 	struct block *l;
 	static const char *empty[] = { null };
 
-	l = (struct block *)alloc(sizeof (struct block), ATEMP);
+	l = alloc(sizeof(struct block), ATEMP);
 	l->flags = 0;
 	ainit(&l->area); /* todo: could use e->area (l->area => l->areap) */
 	if (!e->loc) {
@@ -79,41 +108,38 @@ popblock(void)
 }
 
 /* called by main() to initialise variable data structures */
+#define VARSPEC_DEFNS
+#include "var_spec.h"
+
+enum var_specs {
+#define VARSPEC_ENUMS
+#include "var_spec.h"
+	V_MAX
+};
+
+static const char * const initvar_names[] = {
+#define VARSPEC_ITEMS
+#include "var_spec.h"
+};
+
 void
 initvar(void)
 {
-	static const struct {
-		const char *name;
-		int v;
-	} names[] = {
-		{ "COLUMNS",		V_COLUMNS },
-		{ "IFS",		V_IFS },
-		{ "OPTIND",		V_OPTIND },
-		{ "PATH",		V_PATH },
-		{ "TMPDIR",		V_TMPDIR },
-#if HAVE_PERSISTENT_HISTORY
-		{ "HISTFILE",		V_HISTFILE },
-#endif
-		{ "HISTSIZE",		V_HISTSIZE },
-		{ "RANDOM",		V_RANDOM },
-		{ "SECONDS",		V_SECONDS },
-		{ "TMOUT",		V_TMOUT },
-		{ "LINENO",		V_LINENO },
-		{ NULL,	0 }
-	};
-	int i;
+	int i = 0;
 	struct tbl *tp;
 
-	ktinit(&specials, APERM, 32); /* must be 2^n (currently 17 specials) */
-	for (i = 0; names[i].name; i++) {
-		tp = ktenter(&specials, names[i].name, hash(names[i].name));
+	ktinit(&specials, APERM,
+	    /* must be 80% of 2^n (currently 12 specials) */ 16);
+	while (i < V_MAX - 1) {
+		tp = ktenter(&specials, initvar_names[i],
+		    hash(initvar_names[i]));
 		tp->flag = DEFINED|ISSET;
-		tp->type = names[i].v;
+		tp->type = ++i;
 	}
 }
 
-/* Used to calculate an array index for global()/local().  Sets *arrayp to
- * non-zero if this is an array, sets *valp to the array index, returns
+/* Used to calculate an array index for global()/local(). Sets *arrayp to
+ * true if this is an array, sets *valp to the array index, returns
  * the basename of the array.
  */
 static const char *
@@ -121,24 +147,51 @@ array_index_calc(const char *n, bool *arrayp, uint32_t *valp)
 {
 	const char *p;
 	int len;
+	char *ap = NULL;
 
 	*arrayp = false;
+ redo_from_ref:
 	p = skip_varname(n, false);
+	if (!set_refflag && (p != n) && ksh_isalphx(n[0])) {
+		struct block *l = e->loc;
+		struct tbl *vp;
+		char *vn;
+		uint32_t h;
+
+		strndupx(vn, n, p - n, ATEMP);
+		h = hash(vn);
+		/* check if this is a reference */
+		do {
+			vp = ktsearch(&l->vars, vn, h);
+		} while (!vp && (l = l->next));
+		afree(vn, ATEMP);
+		if (vp && (vp->flag & (DEFINED|ASSOC|ARRAY)) ==
+		    (DEFINED|ASSOC)) {
+			char *cp;
+
+			/* gotcha! */
+			cp = shf_smprintf("%s%s", str_val(vp), p);
+			afree(ap, ATEMP);
+			n = ap = cp;
+			goto redo_from_ref;
+		}
+	}
+
 	if (p != n && *p == '[' && (len = array_ref_len(p))) {
 		char *sub, *tmp;
-		long rval;
+		mksh_ari_t rval;
 
 		/* Calculate the value of the subscript */
 		*arrayp = true;
-		tmp = str_nsave(p + 1, len - 2, ATEMP);
+		strndupx(tmp, p + 1, len - 2, ATEMP);
 		sub = substitute(tmp, 0);
 		afree(tmp, ATEMP);
-		n = str_nsave(n, p - n, ATEMP);
+		strndupx(n, n, p - n, ATEMP);
 		evaluate(sub, &rval, KSH_UNWIND_ERROR, true);
 		*valp = (uint32_t)rval;
 		afree(sub, ATEMP);
 	}
-	return n;
+	return (n);
 }
 
 /*
@@ -150,9 +203,8 @@ global(const char *n)
 	struct block *l = e->loc;
 	struct tbl *vp;
 	int c;
-	unsigned int h;
 	bool array;
-	uint32_t val;
+	uint32_t h, val;
 
 	/* Check to see if this is an array */
 	n = array_index_calc(n, &array, &val);
@@ -173,11 +225,11 @@ global(const char *n)
 				/* setstr can't fail here */
 				setstr(vp, l->argv[c], KSH_RETURN_ERROR);
 			vp->flag |= RDONLY;
-			return vp;
+			return (vp);
 		}
 		vp->flag |= RDONLY;
 		if (n[1] != '\0')
-			return vp;
+			return (vp);
 		vp->flag |= ISSET|INTEGER;
 		switch (c) {
 		case '$':
@@ -201,15 +253,15 @@ global(const char *n)
 		default:
 			vp->flag &= ~(ISSET|INTEGER);
 		}
-		return vp;
+		return (vp);
 	}
 	for (l = e->loc; ; l = l->next) {
 		vp = ktsearch(&l->vars, n, h);
 		if (vp != NULL) {
 			if (array)
-				return arraysearch(vp, val);
+				return (arraysearch(vp, val));
 			else
-				return vp;
+				return (vp);
 		}
 		if (l->next == NULL)
 			break;
@@ -220,7 +272,7 @@ global(const char *n)
 	vp->flag |= DEFINED;
 	if (special(n))
 		vp->flag |= SPECIAL;
-	return vp;
+	return (vp);
 }
 
 /*
@@ -231,9 +283,8 @@ local(const char *n, bool copy)
 {
 	struct block *l = e->loc;
 	struct tbl *vp;
-	unsigned int h;
 	bool array;
-	uint32_t val;
+	uint32_t h, val;
 
 	/* Check to see if this is an array */
 	n = array_index_calc(n, &array, &val);
@@ -243,7 +294,7 @@ local(const char *n, bool copy)
 		vp->flag = DEFINED|RDONLY;
 		vp->type = 0;
 		vp->areap = ATEMP;
-		return vp;
+		return (vp);
 	}
 	vp = ktenter(&l->vars, n, h);
 	if (copy && !(vp->flag & DEFINED)) {
@@ -266,7 +317,7 @@ local(const char *n, bool copy)
 	vp->flag |= DEFINED;
 	if (special(n))
 		vp->flag |= SPECIAL;
-	return vp;
+	return (vp);
 }
 
 /* get variable string value */
@@ -282,18 +333,18 @@ str_val(struct tbl *vp)
 	else if (!(vp->flag&INTEGER))	/* string source */
 		s = vp->val.s + vp->type;
 	else {				/* integer source */
-		/* worst case number length is when base=2, so use BITS(long) */
-		/* minus base #     number    null */
-		char strbuf[1 + 2 + 1 + 8 * sizeof(long) + 1];
+		/* worst case number length is when base=2 */
+		/* 1 (minus) + 2 (base, up to 36) + 1 ('#') + number of bits
+		 * in the mksh_uari_t + 1 (NUL) */
+		char strbuf[1 + 2 + 1 + 8 * sizeof(mksh_uari_t) + 1];
 		const char *digits = (vp->flag & UCASEV_AL) ?
-		    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ" :
-		    "0123456789abcdefghijklmnopqrstuvwxyz";
-		unsigned long n;
+		    digits_uc : digits_lc;
+		mksh_uari_t n;
 		int base;
 
 		s = strbuf + sizeof(strbuf);
 		if (vp->flag & INT_U)
-			n = (unsigned long) vp->val.i;
+			n = vp->val.u;
 		else
 			n = (vp->val.i < 0) ? -vp->val.i : vp->val.i;
 		base = (vp->type == 0) ? 10 : vp->type;
@@ -303,7 +354,8 @@ str_val(struct tbl *vp)
 
 			*(s = strbuf) = '1';
 			s[1] = '#';
-			if (!Flag(FUTFHACK) || ((n & 0xFF80) == 0xEF80))
+			if (!UTFMODE || ((n & 0xFF80) == 0xEF80))
+				/* OPTU-16 -> raw octet */
 				s[2] = n & 0xFF;
 			else
 				sz = utf_wctomb(s + 2, n);
@@ -326,23 +378,23 @@ str_val(struct tbl *vp)
 		if (vp->flag & (RJUST|LJUST)) /* case already dealt with */
 			s = formatstr(vp, s);
 		else
-			s = str_save(s, ATEMP);
+			strdupx(s, s, ATEMP);
 	}
-	return s;
+	return (s);
 }
 
 /* get variable integer value, with error checking */
-long
+static mksh_ari_t
 intval(struct tbl *vp)
 {
-	long num;
+	mksh_ari_t num;
 	int base;
 
 	base = getint(vp, &num, false);
 	if (base == -1)
 		/* XXX check calls - is error here ok by POSIX? */
 		errorf("%s: bad number", str_val(vp));
-	return num;
+	return (num);
 }
 
 /* set variable to string value */
@@ -357,7 +409,7 @@ setstr(struct tbl *vq, const char *s, int error_ok)
 		warningf(true, "%s: is read only", vq->name);
 		if (!error_ok)
 			errorfz();
-		return 0;
+		return (0);
 	}
 	if (!(vq->flag&INTEGER)) { /* string dest */
 		if ((vq->flag&ALLOC)) {
@@ -367,16 +419,16 @@ setstr(struct tbl *vq, const char *s, int error_ok)
 				internal_errorf(
 				    "setstr: %s=%s: assigning to self",
 				    vq->name, s);
-			afree((void*)vq->val.s, vq->areap);
+			afree(vq->val.s, vq->areap);
 		}
 		vq->flag &= ~(ISSET|ALLOC);
 		vq->type = 0;
 		if (s && (vq->flag & (UCASEV_AL|LCASEV|LJUST|RJUST)))
 			s = salloc = formatstr(vq, s);
 		if ((vq->flag&EXPORT))
-			export(vq, s);
+			exportprep(vq, s);
 		else {
-			vq->val.s = str_save(s, vq->areap);
+			strdupx(vq->val.s, s, vq->areap);
 			vq->flag |= ALLOC;
 		}
 	} else {		/* integer dest */
@@ -392,7 +444,7 @@ setstr(struct tbl *vq, const char *s, int error_ok)
 
 /* set variable to integer */
 void
-setint(struct tbl *vq, long int n)
+setint(struct tbl *vq, mksh_ari_t n)
 {
 	if (!(vq->flag&INTEGER)) {
 		struct tbl *vp = &vtemp;
@@ -409,22 +461,22 @@ setint(struct tbl *vq, long int n)
 		setspec(vq);
 }
 
-int
-getint(struct tbl *vp, long int *nump, bool arith)
+static int
+getint(struct tbl *vp, mksh_ari_t *nump, bool arith)
 {
 	char *s;
 	int c, base, neg;
 	bool have_base = false;
-	long num;
+	mksh_ari_t num;
 
 	if (vp->flag&SPECIAL)
 		getspec(vp);
 	/* XXX is it possible for ISSET to be set and val.s to be 0? */
 	if (!(vp->flag&ISSET) || (!(vp->flag&INTEGER) && vp->val.s == NULL))
-		return -1;
+		return (-1);
 	if (vp->flag&INTEGER) {
 		*nump = vp->val.i;
-		return vp->type;
+		return (vp->type);
 	}
 	s = vp->val.s + vp->type;
 	if (s == NULL)	/* redundant given initial test */
@@ -455,11 +507,17 @@ getint(struct tbl *vp, long int *nump, bool arith)
 			if (base == 1) {
 				unsigned int wc;
 
-				if (!Flag(FUTFHACK))
+				if (!UTFMODE)
 					wc = *(unsigned char *)s;
 				else if (utf_mbtowc(&wc, s) == (size_t)-1)
+					/* OPTU-8 -> OPTU-16 */
+					/*
+					 * (with a twist: 1#\uEF80 converts
+					 * the same as 1#\x80 does, thus is
+					 * not round-tripping correctly XXX)
+					 */
 					wc = 0xEF00 + *(unsigned char *)s;
-				*nump = (long)wc;
+				*nump = (mksh_ari_t)wc;
 				return (1);
 			}
 			num = 0;
@@ -472,15 +530,15 @@ getint(struct tbl *vp, long int *nump, bool arith)
 		else if (ksh_isupper(c))
 			c -= 'A' - 10;
 		else
-			return -1;
+			return (-1);
 		if (c < 0 || c >= base)
-			return -1;
+			return (-1);
 		num = num * base + c;
 	}
 	if (neg)
 		num = -num;
 	*nump = num;
-	return base;
+	return (base);
 }
 
 /* convert variable vq to integer variable, setting its value from vp
@@ -490,10 +548,10 @@ struct tbl *
 setint_v(struct tbl *vq, struct tbl *vp, bool arith)
 {
 	int base;
-	long num;
+	mksh_ari_t num;
 
 	if ((base = getint(vp, &num, arith)) == -1)
-		return NULL;
+		return (NULL);
 	if (!(vq->flag & INTEGER) && (vq->flag & ALLOC)) {
 		vq->flag &= ~ALLOC;
 		afree(vq->val.s, vq->areap);
@@ -504,7 +562,7 @@ setint_v(struct tbl *vq, struct tbl *vp, bool arith)
 	vq->flag |= ISSET|INTEGER;
 	if (vq->flag&SPECIAL)
 		setspec(vq);
-	return vq;
+	return (vq);
 }
 
 static char *
@@ -523,7 +581,7 @@ formatstr(struct tbl *vp, const char *s)
 	} else
 		nlen = olen;
 
-	p = (char *)alloc((psiz = nlen * /* MB_LEN_MAX */ 3 + 1), ATEMP);
+	p = alloc((psiz = nlen * /* MB_LEN_MAX */ 3 + 1), ATEMP);
 	if (vp->flag & (RJUST|LJUST)) {
 		int slen = olen, i = 0;
 
@@ -533,7 +591,7 @@ formatstr(struct tbl *vp, const char *s)
 
 			while (i < slen)
 				i += utf_widthadj(qq, &qq);
-			/* strip trailing spaces (at&t uses qq[-1] == ' ') */
+			/* strip trailing spaces (AT&T uses qq[-1] == ' ') */
 			while (qq > s && ksh_isspace(qq[-1])) {
 				--qq;
 				--slen;
@@ -580,14 +638,14 @@ formatstr(struct tbl *vp, const char *s)
 			*q = ksh_tolower(*q);
 	}
 
-	return p;
+	return (p);
 }
 
 /*
  * make vp->val.s be "name=value" for quick exporting.
  */
 static void
-export(struct tbl *vp, const char *val)
+exportprep(struct tbl *vp, const char *val)
 {
 	char *xp;
 	char *op = (vp->flag&ALLOC) ? vp->val.s : NULL;
@@ -595,14 +653,14 @@ export(struct tbl *vp, const char *val)
 	int vallen = strlen(val) + 1;
 
 	vp->flag |= ALLOC;
-	xp = (char*)alloc(namelen + 1 + vallen, vp->areap);
+	xp = alloc(namelen + 1 + vallen, vp->areap);
 	memcpy(vp->val.s = xp, vp->name, namelen);
 	xp += namelen;
 	*xp++ = '=';
 	vp->type = xp - vp->val.s; /* offset to value */
 	memcpy(xp, val, vallen);
 	if (op != NULL)
-		afree((void*)op, vp->areap);
+		afree(op, vp->areap);
 }
 
 /*
@@ -617,19 +675,23 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 	struct tbl *vpbase, *t;
 	char *tvar;
 	const char *val;
+	int len;
 
 	/* check for valid variable name, search for value */
 	val = skip_varname(var, false);
 	if (val == var)
-		return NULL;
+		return (NULL);
+	mkssert(var != NULL);
+	mkssert(*var != 0);
 	if (*val == '[') {
-		int len;
-
+		if (set_refflag)
+			errorf("%s: reference variable cannot be an array",
+			    var);
 		len = array_ref_len(val);
 		if (len == 0)
-			return NULL;
+			return (NULL);
 		/* IMPORT is only used when the shell starts up and is
-		 * setting up its environment.  Allow only simple array
+		 * setting up its environment. Allow only simple array
 		 * references at this time since parameter/command substitution
 		 * is preformed on the [expression] which would be a major
 		 * security hole.
@@ -638,19 +700,23 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 			int i;
 			for (i = 1; i < len - 1; i++)
 				if (!ksh_isdigit(val[i]))
-					return NULL;
+					return (NULL);
 		}
 		val += len;
 	}
-	if (*val == '=') {
-		int i = val++ - var;
-		tvar = str_nsave(var, i, ATEMP);
-	} else {
+	if (*val == '=')
+		strndupx(tvar, var, val++ - var, ATEMP);
+	else {
 		/* Importing from original environment: must have an = */
 		if (set & IMPORT)
-			return NULL;
-		tvar = str_save(var, ATEMP);
+			return (NULL);
+		strdupx(tvar, var, ATEMP);
 		val = NULL;
+		/* handle foo[*] ⇒ foo (whole array) mapping for R39b */
+		len = strlen(tvar);
+		if (len > 3 && tvar[len-3] == '[' && tvar[len-2] == '*' &&
+		    tvar[len-1] == ']')
+			tvar[len-3] = '\0';
 	}
 
 	/* Prevent typeset from creating a local PATH/ENV/SHELL */
@@ -660,11 +726,31 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 
 	vp = (set&LOCAL) ? local(tvar, (set & LOCAL_COPY) ? true : false) :
 	    global(tvar);
+	if (set_refflag == 2 && (vp->flag & (ARRAY|ASSOC)) == ASSOC)
+		vp->flag &= ~ASSOC;
+	else if (set_refflag == 1) {
+		if (vp->flag & ARRAY) {
+			struct tbl *a, *tmp;
+
+			/* Free up entire array */
+			for (a = vp->u.array; a; ) {
+				tmp = a;
+				a = a->u.array;
+				if (tmp->flag & ALLOC)
+					afree(tmp->val.s, tmp->areap);
+				afree(tmp, tmp->areap);
+			}
+			vp->u.array = NULL;
+			vp->flag &= ~ARRAY;
+		}
+		vp->flag |= ASSOC;
+	}
+
 	set &= ~(LOCAL|LOCAL_COPY);
 
 	vpbase = (vp->flag & ARRAY) ? global(arrayname(var)) : vp;
 
-	/* only allow export flag to be set.  at&t ksh allows any attribute to
+	/* only allow export flag to be set. AT&T ksh allows any attribute to
 	 * be changed which means it can be truncated or modified (-L/-R/-Z/-i)
 	 */
 	if ((vpbase->flag&RDONLY) &&
@@ -675,12 +761,13 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 
 	/* most calls are with set/clr == 0 */
 	if (set | clr) {
-		int ok = 1;
+		bool ok = true;
+
 		/* XXX if x[0] isn't set, there will be problems: need to have
 		 * one copy of attributes for arrays...
 		 */
 		for (t = vpbase; t; t = t->u.array) {
-			int fake_assign;
+			bool fake_assign;
 			char *s = NULL;
 			char *free_me = NULL;
 
@@ -717,19 +804,18 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 					 * zap contents of variable, but keep
 					 * the flag settings.
 					 */
-					ok = 0;
+					ok = false;
 					if (t->flag & INTEGER)
 						t->flag &= ~ISSET;
 					else {
 						if (t->flag & ALLOC)
-							afree((void*) t->val.s,
-							    t->areap);
+							afree(t->val.s, t->areap);
 						t->flag &= ~(ISSET|ALLOC);
 						t->type = 0;
 					}
 				}
 				if (free_me)
-					afree((void *) free_me, t->areap);
+					afree(free_me, t->areap);
 			}
 		}
 		if (!ok)
@@ -751,20 +837,22 @@ typeset(const char *var, Tflag set, Tflag clr, int field, int base)
 	/* only x[0] is ever exported, so use vpbase */
 	if ((vpbase->flag&EXPORT) && !(vpbase->flag&INTEGER) &&
 	    vpbase->type == 0)
-		export(vpbase, (vpbase->flag&ISSET) ? vpbase->val.s : null);
+		exportprep(vpbase, (vpbase->flag&ISSET) ? vpbase->val.s : null);
 
-	return vp;
+	return (vp);
 }
 
-/* Unset a variable.  array_ref is set if there was an array reference in
- * the name lookup (eg, x[2]).
+/**
+ * Unset a variable. The flags can be:
+ * |1	= tear down entire array
+ * |2	= keep attributes, only unset content
  */
 void
-unset(struct tbl *vp, int array_ref)
+unset(struct tbl *vp, int flags)
 {
 	if (vp->flag & ALLOC)
-		afree((void*)vp->val.s, vp->areap);
-	if ((vp->flag & ARRAY) && !array_ref) {
+		afree(vp->val.s, vp->areap);
+	if ((vp->flag & ARRAY) && (flags & 1)) {
 		struct tbl *a, *tmp;
 
 		/* Free up entire array */
@@ -772,20 +860,24 @@ unset(struct tbl *vp, int array_ref)
 			tmp = a;
 			a = a->u.array;
 			if (tmp->flag & ALLOC)
-				afree((void *) tmp->val.s, tmp->areap);
+				afree(tmp->val.s, tmp->areap);
 			afree(tmp, tmp->areap);
 		}
 		vp->u.array = NULL;
 	}
+	if (flags & 2) {
+		vp->flag &= ~(ALLOC|ISSET);
+		return;
+	}
 	/* If foo[0] is being unset, the remainder of the array is kept... */
-	vp->flag &= SPECIAL | (array_ref ? ARRAY|DEFINED : 0);
+	vp->flag &= SPECIAL | ((flags & 1) ? 0 : ARRAY|DEFINED);
 	if (vp->flag & SPECIAL)
 		unsetspec(vp);	/* responsible for 'unspecial'ing var */
 }
 
 /* return a pointer to the first char past a legal variable name (returns the
- * argument if there is no legal name, returns * a pointer to the terminating
- * null if whole string is legal).
+ * argument if there is no legal name, returns a pointer to the terminating
+ * NUL if whole string is legal).
  */
 const char *
 skip_varname(const char *s, int aok)
@@ -801,7 +893,7 @@ skip_varname(const char *s, int aok)
 	return (s);
 }
 
-/* Return a pointer to the first character past any legal variable name.  */
+/* Return a pointer to the first character past any legal variable name */
 const char *
 skip_wdvarname(const char *s,
     int aok)				/* skip array de-reference? */
@@ -839,7 +931,7 @@ is_wdvarname(const char *s, int aok)
 {
 	const char *p = skip_wdvarname(s, aok);
 
-	return p != s && p[0] == EOS;
+	return (p != s && p[0] == EOS);
 }
 
 /* Check if coded string s is a variable assignment */
@@ -848,7 +940,7 @@ is_wdvarassign(const char *s)
 {
 	const char *p = skip_wdvarname(s, true);
 
-	return p != s && p[0] == CHAR && p[1] == '=';
+	return (p != s && p[0] == CHAR && p[1] == '=');
 }
 
 /*
@@ -858,18 +950,18 @@ char **
 makenv(void)
 {
 	struct block *l;
-	XPtrV env;
+	XPtrV denv;
 	struct tbl *vp, **vpp;
 	int i;
 
-	XPinit(env, 64);
+	XPinit(denv, 64);
 	for (l = e->loc; l != NULL; l = l->next)
 		for (vpp = l->vars.tbls, i = l->vars.size; --i >= 0; )
 			if ((vp = *vpp++) != NULL &&
 			    (vp->flag&(ISSET|EXPORT)) == (ISSET|EXPORT)) {
 				struct block *l2;
 				struct tbl *vp2;
-				unsigned int h = hash(vp->name);
+				uint32_t h = hash(vp->name);
 
 				/* unexport any redefined instances */
 				for (l2 = l->next; l2 != NULL; l2 = l2->next) {
@@ -881,106 +973,94 @@ makenv(void)
 					/* integer to string */
 					char *val;
 					val = str_val(vp);
-					vp->flag &= ~(INTEGER|RDONLY);
+					vp->flag &= ~(INTEGER|RDONLY|SPECIAL);
 					/* setstr can't fail here */
 					setstr(vp, val, KSH_RETURN_ERROR);
 				}
-				XPput(env, vp->val.s);
+				XPput(denv, vp->val.s);
 			}
-	XPput(env, NULL);
-	return (char **) XPclose(env);
+	XPput(denv, NULL);
+	return ((char **)XPclose(denv));
 }
 
-/*
- * Get us a random number, either from rand(3) or arc4random(3), with
- * the latter being preferred. If Flag(FARC4RANDOM) is 0, we use rand(3),
- * otherwise arc4random(3). We have static caches to make change_random
- * and writes to $RANDOM a cheap operation.
- */
-#if HAVE_ARC4RANDOM
-static uint32_t rnd_cache[2];
-static char rnd_lastflag = 2;
-#endif
-
-static int
-rnd_get(void)
+/* Bob Jenkins' one-at-a-time hash */
+static uint32_t
+oaathash_update(register uint32_t h, register const uint8_t *cp,
+    register size_t n)
 {
-#if HAVE_ARC4RANDOM
-#if HAVE_ARC4RANDOM_PUSHB
-	uint32_t rv = 0;
-#endif
-	if (Flag(FARC4RANDOM) != rnd_lastflag) {
-		if (Flag(FARC4RANDOM) == 0) {
-			/* transition to 0 by set: srand */
-			srand(arc4random() & 0x7FFF);
-		} else if (rnd_lastflag == 0) {
-			/* transition from 0: addrandom */
-			rnd_cache[0] ^= rand();
-			rnd_cache[1] ^= rand();
-		}
-		rnd_lastflag = Flag(FARC4RANDOM);
+	while (n--) {
+		h += *cp++;
+		h += h << 10;
+		h ^= h >> 6;
 	}
-	if (Flag(FARC4RANDOM)) {
-		if (rnd_cache[0] || rnd_cache[1])
-#if HAVE_ARC4RANDOM_PUSHB
-			rv = arc4random_pushb(rnd_cache, sizeof (rnd_cache));
-#else
-			arc4random_addrandom((void *)rnd_cache,
-			    sizeof (rnd_cache));
-#endif
-		rnd_cache[0] = rnd_cache[1] = 0;
-		return ((
-#if HAVE_ARC4RANDOM_PUSHB
-		    rv ? rv :
-#endif
-		    arc4random()) & 0x7FFF);
-	}
-#endif
-	return (rand() & 0x7FFF);
+
+	return (h);
 }
 
-static void
-rnd_set(unsigned long newval)
+static uint32_t
+oaathash_finalise(register uint32_t h)
 {
-#if HAVE_ARC4RANDOM
-	rnd_cache[0] ^= (newval << 15) | rand();
-	rnd_cache[1] ^= newval >> 17;
-	if (Flag(FARC4RANDOM) == 1)
-		return;
-	if (Flag(FARC4RANDOM) == 2)
-		Flag(FARC4RANDOM) = 0;
-	/* transition to 0 by write: only srand */
-	rnd_lastflag = 0;
-#endif
-	srand(newval & 0x7FFF);
+	h += h << 3;
+	h ^= h >> 11;
+	h += h << 15;
+
+	return (h);
+}
+
+uint32_t
+oaathash_full(register const uint8_t *bp)
+{
+	register uint32_t h = 0;
+	register uint8_t c;
+
+	while ((c = *bp++)) {
+		h += c;
+		h += h << 10;
+		h ^= h >> 6;
+	}
+
+	return (oaathash_finalise(h));
 }
 
-/*
- * Called after a fork in parent to bump the random number generator.
- * Done to ensure children will not get the same random number sequence
- * if the parent doesn't use $RANDOM.
- */
 void
-change_random(unsigned long newval)
+change_random(const void *vp, size_t n)
 {
-	int rval = 0;
-
-#if HAVE_ARC4RANDOM
-	if (Flag(FARC4RANDOM)) {
-		rnd_cache[0] ^= (newval << 15) | rand();
-		rnd_cache[1] ^= newval >> 17;
-		return;
-	}
+	register uint32_t h = 0x100;
+#if defined(__OpenBSD__)
+	int mib[2];
+	uint8_t k[3];
+	size_t klen;
 #endif
 
-	rval += newval & 0x7FFF;
-	newval >>= 15;
-	rval += newval & 0x7FFF;
-	newval >>= 15;
-	rval += newval + rand();
-	rval = (rval & 0x7FFF) ^ (rval >> 15);
+	kshstate_v.cr_dp = vp;
+	kshstate_v.cr_dsz = n;
+	gettimeofday(&kshstate_v.cr_tv, NULL);
+	h = oaathash_update(oaathash_update(h, (void *)&kshstate_v,
+	    sizeof(kshstate_v)), vp, n);
+	kshstate_v.lcg_state_ = oaathash_finalise(h);
 
-	srand(rval);
+#if defined(__OpenBSD__)
+	/* OpenBSD, MirBSD: proper kernel entropy comes at zero cost */
+
+	mib[0] = CTL_KERN;
+	mib[1] = KERN_ARND;
+	klen = sizeof(k);
+	sysctl(mib, 2, k, &klen, &kshstate_v.lcg_state_,
+	    sizeof(kshstate_v.lcg_state_));
+	/* we ignore failures and take in k anyway */
+	h = oaathash_update(h, k, sizeof(k));
+	kshstate_v.lcg_state_ = oaathash_finalise(h);
+#elif defined(MKSH_A4PB)
+	/* forced by the user to use arc4random_pushb(3) • Cygwin? */
+	{
+		uint32_t prv;
+
+		prv = arc4random_pushb(&kshstate_v.lcg_state_,
+		    sizeof(kshstate_v.lcg_state_));
+		h = oaathash_update(h, &prv, sizeof(prv));
+	}
+	kshstate_v.lcg_state_ = oaathash_finalise(h);
+#endif
 }
 
 /*
@@ -994,7 +1074,7 @@ special(const char *name)
 	struct tbl *tp;
 
 	tp = ktsearch(&specials, name, hash(name));
-	return tp && (tp->flag & ISSET) ? tp->type : V_NONE;
+	return (tp && (tp->flag & ISSET) ? tp->type : V_NONE);
 }
 
 /* Make a variable non-special */
@@ -1008,72 +1088,87 @@ unspecial(const char *name)
 		ktdelete(tp);
 }
 
-static	time_t	seconds;		/* time SECONDS last set */
-static	int	user_lineno;		/* what user set $LINENO to */
+static time_t seconds;		/* time SECONDS last set */
+static int user_lineno;		/* what user set $LINENO to */
 
 static void
 getspec(struct tbl *vp)
 {
-	switch (special(vp->name)) {
+	register mksh_ari_t i;
+	int st;
+
+	switch ((st = special(vp->name))) {
 	case V_SECONDS:
-		vp->flag &= ~SPECIAL;
-		/* On start up the value of SECONDS is used before seconds
-		 * has been set - don't do anything in this case
+		/*
+		 * On start up the value of SECONDS is used before
+		 * it has been set - don't do anything in this case
 		 * (see initcoms[] in main.c).
 		 */
 		if (vp->flag & ISSET) {
 			struct timeval tv;
 
 			gettimeofday(&tv, NULL);
-			setint(vp, tv.tv_sec - seconds);
-		}
-		vp->flag |= SPECIAL;
+			i = tv.tv_sec - seconds;
+		} else
+			return;
 		break;
 	case V_RANDOM:
-		vp->flag &= ~SPECIAL;
-		setint(vp, rnd_get());
-		vp->flag |= SPECIAL;
+		/*
+		 * this is the same Linear Congruential PRNG as Borland
+		 * C/C++ allegedly uses in its built-in rand() function
+		 */
+		i = ((kshstate_v.lcg_state_ =
+		    22695477 * kshstate_v.lcg_state_ + 1) >> 16) & 0x7FFF;
 		break;
 	case V_HISTSIZE:
-		vp->flag &= ~SPECIAL;
-		setint(vp, (long) histsize);
-		vp->flag |= SPECIAL;
+		i = histsize;
 		break;
 	case V_OPTIND:
-		vp->flag &= ~SPECIAL;
-		setint(vp, (long) user_opt.uoptind);
-		vp->flag |= SPECIAL;
+		i = user_opt.uoptind;
 		break;
 	case V_LINENO:
-		vp->flag &= ~SPECIAL;
-		setint(vp, (long) current_lineno + user_lineno);
-		vp->flag |= SPECIAL;
+		i = current_lineno + user_lineno;
 		break;
+	case V_COLUMNS:
+	case V_LINES:
+		/*
+		 * Do NOT export COLUMNS/LINES. Many applications
+		 * check COLUMNS/LINES before checking ws.ws_col/row,
+		 * so if the app is started with C/L in the environ
+		 * and the window is then resized, the app won't
+		 * see the change cause the environ doesn't change.
+		 */
+		change_winsz();
+		i = st == V_COLUMNS ? x_cols : x_lins;
+		break;
+	default:
+		/* do nothing, do not touch vp at all */
+		return;
 	}
+	vp->flag &= ~SPECIAL;
+	setint(vp, i);
+	vp->flag |= SPECIAL;
 }
 
 static void
 setspec(struct tbl *vp)
 {
+	mksh_ari_t i;
 	char *s;
+	int st;
 
-	switch (special(vp->name)) {
+	switch ((st = special(vp->name))) {
 	case V_PATH:
 		if (path)
 			afree(path, APERM);
 		s = str_val(vp);
-		path = str_save(s, APERM);
+		strdupx(path, s, APERM);
 		flushcom(1);	/* clear tracked aliases */
-		break;
+		return;
 	case V_IFS:
 		setctypes(s = str_val(vp), C_IFS);
 		ifs0 = *s;
-		break;
-	case V_OPTIND:
-		vp->flag &= ~SPECIAL;
-		getopts_reset((int)intval(vp));
-		vp->flag |= SPECIAL;
-		break;
+		return;
 	case V_TMPDIR:
 		if (tmpdir) {
 			afree(tmpdir, APERM);
@@ -1088,48 +1183,72 @@ setspec(struct tbl *vp)
 			s = str_val(vp);
 			if (s[0] == '/' && access(s, W_OK|X_OK) == 0 &&
 			    stat(s, &statb) == 0 && S_ISDIR(statb.st_mode))
-				tmpdir = str_save(s, APERM);
+				strdupx(tmpdir, s, APERM);
 		}
 		break;
-	case V_HISTSIZE:
-		vp->flag &= ~SPECIAL;
-		sethistsize((int)intval(vp));
-		vp->flag |= SPECIAL;
-		break;
 #if HAVE_PERSISTENT_HISTORY
 	case V_HISTFILE:
 		sethistfile(str_val(vp));
 		break;
 #endif
-	case V_COLUMNS:
-		if ((x_cols = intval(vp)) <= MIN_COLS)
-			x_cols = MIN_COLS;
+	case V_TMOUT:
+		/* AT&T ksh seems to do this (only listen if integer) */
+		if (vp->flag & INTEGER)
+			ksh_tmout = vp->val.i >= 0 ? vp->val.i : 0;
 		break;
+
+	/* common sub-cases */
+	case V_OPTIND:
+	case V_HISTSIZE:
+	case V_COLUMNS:
+	case V_LINES:
 	case V_RANDOM:
+	case V_SECONDS:
+	case V_LINENO:
 		vp->flag &= ~SPECIAL;
-		rnd_set(intval(vp));
+		i = intval(vp);
 		vp->flag |= SPECIAL;
 		break;
+	default:
+		/* do nothing, do not touch vp at all */
+		return;
+	}
+
+	/* process the singular parts of the common cases */
+
+	switch (st) {
+	case V_OPTIND:
+		getopts_reset((int)i);
+		break;
+	case V_HISTSIZE:
+		sethistsize((int)i);
+		break;
+	case V_COLUMNS:
+		if (i >= MIN_COLS)
+			x_cols = i;
+		break;
+	case V_LINES:
+		if (i >= MIN_LINS)
+			x_lins = i;
+		break;
+	case V_RANDOM:
+		/*
+		 * mksh R39d+ no longer has the traditional repeatability
+		 * of $RANDOM sequences, but always retains state
+		 */
+		change_random(&i, sizeof(i));
+		break;
 	case V_SECONDS:
-		vp->flag &= ~SPECIAL;
 		{
 			struct timeval tv;
 
 			gettimeofday(&tv, NULL);
-			seconds = tv.tv_sec - intval(vp);
+			seconds = tv.tv_sec - i;
 		}
-		vp->flag |= SPECIAL;
-		break;
-	case V_TMOUT:
-		/* at&t ksh seems to do this (only listen if integer) */
-		if (vp->flag & INTEGER)
-			ksh_tmout = vp->val.i >= 0 ? vp->val.i : 0;
 		break;
 	case V_LINENO:
-		vp->flag &= ~SPECIAL;
 		/* The -1 is because line numbering starts at 1. */
-		user_lineno = (unsigned int)intval(vp) - current_lineno - 1;
-		vp->flag |= SPECIAL;
+		user_lineno = (unsigned int)i - current_lineno - 1;
 		break;
 	}
 }
@@ -1141,7 +1260,7 @@ unsetspec(struct tbl *vp)
 	case V_PATH:
 		if (path)
 			afree(path, APERM);
-		path = str_save(def_path, APERM);
+		strdupx(path, def_path, APERM);
 		flushcom(1);	/* clear tracked aliases */
 		break;
 	case V_IFS:
@@ -1158,18 +1277,17 @@ unsetspec(struct tbl *vp)
 	case V_LINENO:
 	case V_RANDOM:
 	case V_SECONDS:
-	case V_TMOUT:		/* at&t ksh leaves previous value in place */
+	case V_TMOUT:		/* AT&T ksh leaves previous value in place */
 		unspecial(vp->name);
 		break;
 
-	  /* at&t ksh man page says OPTIND, OPTARG and _ lose special meaning,
-	   * but OPTARG does not (still set by getopts) and _ is also still
-	   * set in various places.
-	   * Don't know what at&t does for:
-	   *		HISTSIZE, HISTFILE,
-	   * Unsetting these in at&t ksh does not loose the 'specialness':
-	   *    no effect: IFS, COLUMNS, PATH, TMPDIR
-	   */
+	/*
+	 * AT&T ksh man page says OPTIND, OPTARG and _ lose special
+	 * meaning, but OPTARG does not (still set by getopts) and _ is
+	 * also still set in various places. Don't know what AT&T does
+	 * for HISTSIZE, HISTFILE. Unsetting these in AT&T ksh does not
+	 * loose the 'specialness': IFS, COLUMNS, PATH, TMPDIR
+	 */
 	}
 }
 
@@ -1180,43 +1298,45 @@ unsetspec(struct tbl *vp)
 static struct tbl *
 arraysearch(struct tbl *vp, uint32_t val)
 {
-	struct tbl *prev, *curr, *new;
-	size_t namelen = strlen(vp->name) + 1;
+	struct tbl *prev, *curr, *news;
+	size_t len;
 
-	vp->flag |= ARRAY|DEFINED;
-	vp->index = 0;
+	vp->flag = (vp->flag | (ARRAY|DEFINED)) & ~ASSOC;
 	/* The table entry is always [0] */
 	if (val == 0)
-		return vp;
+		return (vp);
 	prev = vp;
 	curr = vp->u.array;
-	while (curr && curr->index < val) {
+	while (curr && curr->ua.index < val) {
 		prev = curr;
 		curr = curr->u.array;
 	}
-	if (curr && curr->index == val) {
+	if (curr && curr->ua.index == val) {
 		if (curr->flag&ISSET)
-			return curr;
-		else
-			new = curr;
+			return (curr);
+		news = curr;
 	} else
-		new = (struct tbl *)alloc(sizeof (struct tbl) + namelen,
-		    vp->areap);
-	strlcpy(new->name, vp->name, namelen);
-	new->flag = vp->flag & ~(ALLOC|DEFINED|ISSET|SPECIAL);
-	new->type = vp->type;
-	new->areap = vp->areap;
-	new->u2.field = vp->u2.field;
-	new->index = val;
-	if (curr != new) {		/* not reusing old array entry */
-		prev->u.array = new;
-		new->u.array = curr;
+		news = NULL;
+	len = strlen(vp->name) + 1;
+	if (!news) {
+		news = alloc(offsetof(struct tbl, name[0]) + len, vp->areap);
+		memcpy(news->name, vp->name, len);
 	}
-	return new;
+	news->flag = (vp->flag & ~(ALLOC|DEFINED|ISSET|SPECIAL)) | AINDEX;
+	news->type = vp->type;
+	news->areap = vp->areap;
+	news->u2.field = vp->u2.field;
+	news->ua.index = val;
+
+	if (curr != news) {		/* not reusing old array entry */
+		prev->u.array = news;
+		news->u.array = curr;
+	}
+	return (news);
 }
 
 /* Return the length of an array reference (eg, [1+2]) - cp is assumed
- * to point to the open bracket.  Returns 0 if there is no matching closing
+ * to point to the open bracket. Returns 0 if there is no matching closing
  * bracket.
  */
 int
@@ -1230,8 +1350,8 @@ array_ref_len(const char *cp)
 		if (c == '[')
 			depth++;
 	if (!c)
-		return 0;
-	return s - cp;
+		return (0);
+	return (s - cp);
 }
 
 /*
@@ -1241,39 +1361,130 @@ char *
 arrayname(const char *str)
 {
 	const char *p;
+	char *rv;
 
 	if ((p = cstrchr(str, '[')) == 0)
 		/* Shouldn't happen, but why worry? */
-		return str_save(str, ATEMP);
+		strdupx(rv, str, ATEMP);
+	else
+		strndupx(rv, str, p - str, ATEMP);
 
-	return str_nsave(str, p - str, ATEMP);
+	return (rv);
 }
 
-/* Set (or overwrite, if !reset) the array variable var to the values in vals.
- */
-void
-set_array(const char *var, int reset, const char **vals)
+/* set (or overwrite, if reset) the array variable var to the values in vals */
+mksh_uari_t
+set_array(const char *var, bool reset, const char **vals)
 {
 	struct tbl *vp, *vq;
-	uint32_t i;
+	mksh_uari_t i;
+	const char *ccp;
+#ifndef MKSH_SMALL
+	char *cp;
+	mksh_uari_t j;
+#endif
 
 	/* to get local array, use "typeset foo; set -A foo" */
 	vp = global(var);
 
-	/* Note: at&t ksh allows set -A but not set +A of a read-only var */
+	/* Note: AT&T ksh allows set -A but not set +A of a read-only var */
 	if ((vp->flag&RDONLY))
 		errorf("%s: is read only", var);
 	/* This code is quite non-optimal */
-	if (reset > 0)
+	if (reset)
 		/* trash existing values and attributes */
-		unset(vp, 0);
+		unset(vp, 1);
 	/* todo: would be nice for assignment to completely succeed or
-	 * completely fail.  Only really effects integer arrays:
+	 * completely fail. Only really effects integer arrays:
 	 * evaluation of some of vals[] may fail...
 	 */
-	for (i = 0; vals[i]; i++) {
-		vq = arraysearch(vp, i);
+	i = 0;
+#ifndef MKSH_SMALL
+	j = 0;
+#else
+#define j i
+#endif
+	while ((ccp = vals[i])) {
+#ifndef MKSH_SMALL
+		if (*ccp == '[') {
+			int level = 0;
+
+			while (*ccp) {
+				if (*ccp == ']' && --level == 0)
+					break;
+				if (*ccp == '[')
+					++level;
+				++ccp;
+			}
+			if (*ccp == ']' && level == 0 && ccp[1] == '=') {
+				strndupx(cp, vals[i] + 1, ccp - (vals[i] + 1),
+				    ATEMP);
+				evaluate(substitute(cp, 0), (mksh_ari_t *)&j,
+				    KSH_UNWIND_ERROR, true);
+				afree(cp, ATEMP);
+				ccp += 2;
+			} else
+				ccp = vals[i];
+		}
+#endif
+
+		vq = arraysearch(vp, j);
 		/* would be nice to deal with errors here... (see above) */
-		setstr(vq, vals[i], KSH_RETURN_ERROR);
+		setstr(vq, ccp, KSH_RETURN_ERROR);
+		i++;
+#ifndef MKSH_SMALL
+		j++;
+#endif
 	}
+
+	return (i);
+}
+
+void
+change_winsz(void)
+{
+	if (x_lins < 0) {
+		/* first time initialisation */
+#ifdef TIOCGWINSZ
+		if (tty_fd < 0)
+			/* non-FTALKING, try to get an fd anyway */
+			tty_init(false, false);
+#endif
+		x_cols = -1;
+	}
+
+#ifdef TIOCGWINSZ
+	/* check if window size has changed since first time */
+	if (tty_fd >= 0) {
+		struct winsize ws;
+
+		if (ioctl(tty_fd, TIOCGWINSZ, &ws) >= 0) {
+			if (ws.ws_col)
+				x_cols = ws.ws_col;
+			if (ws.ws_row)
+				x_lins = ws.ws_row;
+		}
+	}
+#endif
+
+	/* bounds check for sane values, use defaults otherwise */
+	if (x_cols < MIN_COLS)
+		x_cols = 80;
+	if (x_lins < MIN_LINS)
+		x_lins = 24;
+
+#ifdef SIGWINCH
+	got_winch = 0;
+#endif
+}
+
+uint32_t
+evilhash(const char *s)
+{
+	register uint32_t h = 0x100;
+
+	h = oaathash_update(h, (void *)&kshstate_f, sizeof(kshstate_f));
+	kshstate_f.h = oaathash_full((const uint8_t *)s);
+	return (oaathash_finalise(oaathash_update(h,
+	    (void *)&kshstate_f.h, sizeof(kshstate_f.h))));
 }
